/* Copyright (C) 2012 Kristian Lauszus, TKJ Electronics. All rights reserved.

   This software may be distributed and modified under the terms of the GNU
   General Public License version 2 (GPL2) as published by the Free Software
   Foundation and appearing in the file GPL2.TXT included in the packaging of
   this file. Please note that GPL2 Section 2[b] requires that all works based
   on this software must also be made publicly available under the terms of
   the GPL2 ("Copyleft").

   Contact information
   -------------------

   Kristian Lauszus, TKJ Electronics
Web      :  http://www.tkjelectronics.com
e-mail   :  kristianl@tkjelectronics.com
*/

#include "BTD.h"
#include "SPP.h"

#define TRACE_USBHOST_HCI(x)

const uint32_t BTD::BTD_CONTROL_PIPE = 0;
const uint32_t BTD::BTD_EVENT_PIPE   = 1;
const uint32_t BTD::BTD_DATAIN_PIPE  = 2;
const uint32_t BTD::BTD_DATAOUT_PIPE = 3;

static BTD * btd_instance = NULL;

void L2CAP_Command_handler(usb_add_t add,
    usb_ep_t ep,
    uhd_trans_status_t status,
    iram_size_t nb_transferred,
    uint8_t callback_data );

// #define KEY_REQUEST_IMPLEMENTED_SAFER

const char * l2cap_message_types[] = {
  "L2CAP_QUEUE_NOOP", 

  /* Item represents parameters for an L2CAP information response */
  "L2CAP_QUEUE_INFORMATION_RESPONSE",  // pBtd->l2cap_information_response

  /* Item represents parameters for an L2CAP information request */
  "L2CAP_QUEUE_INFORMATION_REQUEST",  // pBtd->l2cap_information_request

  "L2CAP_SERVICE_NOT_SUPPORTED",  // 8 pBtd->serviceNotSupported
  "L2CAP_CONNECTION_RESPONSE",  // 9 pBtd->l2cap_connection_response
  "L2CAP_CONFIG_REQUEST",  // 10 pBtd->l2cap_config_request
  "L2CAP_CONFIG_RESPONSE",  // 11 pBtd->l2cap_config_response
  "L2CAP_DISCONNECT_REQUEST",  // 12 pBtd->l2cap_disconnection_request
  "L2CAP_DISCONNECT_RESP",  // 13 pBtd->l2cap_disconnection_response

  "L2CAP_SDP_SERVICE_ATTRIBUTE_RESPONSE_PBR", // 14

  /* RFCOMM_task */
  "RFCOMM_PROTOCOL_MESSAGE",
  "RFCOMM_PROTOCOL_CREDIT",
};


/*** Extended Inquiry Response. See CoreV4, Vol. 3.C.18, Generic Access Profile ***/
const uint8_t arduino_due_btserial[] PROGMEM = {
  
  // EIR Data (Data types as described in https://www.bluetooth.org/en-us/specification/assigned-numbers/generic-access-profile)
  // Length, Type, 
  // Complete local name - EIR type 0x09
  29, 0x09, 'A' , 'r', 'd', 'u', 'i', 'n', 'o', ' ', 'D', 'u', 'e', ' ', 'S', 'e', 'r', 'i', 'a', 'l', ' ', 'B', 'T', ' ', 'S', 'k', 'e', 't', 'c', 'h',
  // Complete List of 16-bit Service UUIDs - 0x03
  5, 0x03,
    0x00, 0x10, // ServiceDiscoveryServer - 0x1000
    0x01, 0x11, // SerialPort - 0x1101
  0x00 // End of Data
};

const uint8_t sdp_pdu_size_mask[] PROGMEM = {
  0x01, // b0000 0001  SDP_PDU_NIL = 0,
  0x1F, // b0001 1111  SDP_PDU_UINT = 1,
  0x1F, // b0001 1111  SDP_PDU_INT = 2,
  0x16, // b0001 0110  SDP_PDU_UUID = 3,
  0xE0, // b1110 0000  SDP_PDU_STR = 4,
  0x01, // b0000 0001  SDP_PDU_BOOL = 5,
  0xE0, // b1110 0000  SDP_PDU_DATASEQ = 6,
  0xE0, // b1110 0000  SDP_PDU_DATAALT = 7,
  0xE0, // b1110 0000  SDP_PDU_URL = 8,
  0x01  //             SDP_PDU_CONTAINER_END = 9,
};

const uint8_t link_key_generator[16] PROGMEM = {
	0x00, 0x91, 0xE3, 0x72, 0x07, 0x96, 0xE4, 0x75,
  0x0E, 0x9F, 0xED, 0x7C, 0x09, 0x98, 0xEA, 0x7B
};

/*
 * CRC (reversed crc) lookup table as calculated by the table generator in ETSI TS 101 369 V6.3.0.
 */
const uint8_t rfcomm_crc_table[256] PROGMEM = {/* reversed, 8-bit, poly=0x07 */
	0x00, 0x91, 0xE3, 0x72, 0x07, 0x96, 0xE4, 0x75, 0x0E, 0x9F, 0xED, 0x7C, 0x09, 0x98, 0xEA, 0x7B,
	0x1C, 0x8D, 0xFF, 0x6E, 0x1B, 0x8A, 0xF8, 0x69, 0x12, 0x83, 0xF1, 0x60, 0x15, 0x84, 0xF6, 0x67,
	0x38, 0xA9, 0xDB, 0x4A, 0x3F, 0xAE, 0xDC, 0x4D, 0x36, 0xA7, 0xD5, 0x44, 0x31, 0xA0, 0xD2, 0x43,
	0x24, 0xB5, 0xC7, 0x56, 0x23, 0xB2, 0xC0, 0x51, 0x2A, 0xBB, 0xC9, 0x58, 0x2D, 0xBC, 0xCE, 0x5F,
	0x70, 0xE1, 0x93, 0x02, 0x77, 0xE6, 0x94, 0x05, 0x7E, 0xEF, 0x9D, 0x0C, 0x79, 0xE8, 0x9A, 0x0B,
	0x6C, 0xFD, 0x8F, 0x1E, 0x6B, 0xFA, 0x88, 0x19, 0x62, 0xF3, 0x81, 0x10, 0x65, 0xF4, 0x86, 0x17,
	0x48, 0xD9, 0xAB, 0x3A, 0x4F, 0xDE, 0xAC, 0x3D, 0x46, 0xD7, 0xA5, 0x34, 0x41, 0xD0, 0xA2, 0x33,
	0x54, 0xC5, 0xB7, 0x26, 0x53, 0xC2, 0xB0, 0x21, 0x5A, 0xCB, 0xB9, 0x28, 0x5D, 0xCC, 0xBE, 0x2F,
	0xE0, 0x71, 0x03, 0x92, 0xE7, 0x76, 0x04, 0x95, 0xEE, 0x7F, 0x0D, 0x9C, 0xE9, 0x78, 0x0A, 0x9B,
	0xFC, 0x6D, 0x1F, 0x8E, 0xFB, 0x6A, 0x18, 0x89, 0xF2, 0x63, 0x11, 0x80, 0xF5, 0x64, 0x16, 0x87,
	0xD8, 0x49, 0x3B, 0xAA, 0xDF, 0x4E, 0x3C, 0xAD, 0xD6, 0x47, 0x35, 0xA4, 0xD1, 0x40, 0x32, 0xA3,
	0xC4, 0x55, 0x27, 0xB6, 0xC3, 0x52, 0x20, 0xB1, 0xCA, 0x5B, 0x29, 0xB8, 0xCD, 0x5C, 0x2E, 0xBF,
	0x90, 0x01, 0x73, 0xE2, 0x97, 0x06, 0x74, 0xE5, 0x9E, 0x0F, 0x7D, 0xEC, 0x99, 0x08, 0x7A, 0xEB,
	0x8C, 0x1D, 0x6F, 0xFE, 0x8B, 0x1A, 0x68, 0xF9, 0x82, 0x13, 0x61, 0xF0, 0x85, 0x14, 0x66, 0xF7,
	0xA8, 0x39, 0x4B, 0xDA, 0xAF, 0x3E, 0x4C, 0xDD, 0xA6, 0x37, 0x45, 0xD4, 0xA1, 0x30, 0x42, 0xD3,
	0xB4, 0x25, 0x57, 0xC6, 0xB3, 0x22, 0x50, 0xC1, 0xBA, 0x2B, 0x59, 0xC8, 0xBD, 0x2C, 0x5E, 0xCF
};

const uint8_t SDP_search_attr_response_pnp[] PROGMEM = {
  // Sample response to a PNP Information record request:
  // 00000000: 9100 4000 0700 0200 8c00 8935 8735 8509  ..@........5.5..
  // 00000000: 9100 4000 07                             SDP Response
  // 00000000:             00 02                        Transaction ID
  //                            00 8c                   Length
  //                                 00 89              Attr List Bytecount
  //                                      35 87         D.E.S., len 0x87
  //                                           35 85    D.E.S., len 0x85
  //                                                09  Type 1,1: UINT16 
  // 00000010: 0000 0a4f 4912 0009 0001 3503 1912 0009  ...OI.....5.....
  // 00000010: 0000                                     0x0000 AttrID: Service Record Handle 
  // 00000010:      0a4f 4912 00                        Service Record Handle Value
  // 00000010:                  09 0001                 0x0001 AttrID: Service Class ID List 
  // 00000010:                          3503 1912 00    D.E.S., 0x03B 
  // 00000010:                               1912 00    UUID16: 0x1200 PnP Information 
  // 00000010:                                      09  Type 1,1: UINT16 
  // 00000020: 0002 0a00 0000 0009 0005 3503 1910 0209  ..........5.....
  // 00000020: 0002                                     0x0002 AttrID: Service Record State 
  // 00000020:      0a00 0000 00                        Service Record State
  // 00000020:                  09 0005                 0x0005 AttrID: Browse Group List 
  // 00000020:                          3503 1910 02    D.E.S., 0x03B, Public Browse Root
  // 00000020:                                      09  0x0002 AttrID:  
  // 00000030: 0006                                     Language Base Attribute ID List
  // 00000030:      3524                                D.E.S., 0x24B (36d bytes)
  // 00000030:           0965 6e09 006a 0901 0009 6672  12B 
  // 00000030: 0006 3524 0965 6e09 006a 0901 0009 6672  ..5$.en..j....fr
  // 00000040: 0900 6a09 0110 0964 6509 006a 0901 2009  16B 
  // 00000050: 6a61 0900 6a09 0130 0900 0808 ff09 0101  ja..j..0........
  // 00000050: 6a61 0900 6a09 0130                      8B 
  // 00000050:                     0900 08              0x0008 AttrID: Service Availability
  // 00000050:                            08 ff         UINT8: 0xFF         
  // 00000050:                                 09 0101  0x0101 AttrID: Reserved / Undecoded 
  // 00000060: 250f 506e 5020 496e 666f 726d 6174 696f  %.PnP Informatio
  // 00000060: 250f                                     Type 4,5: String, 15B 
  // 00000060:      506e 5020 496e 666f 726d 6174 696f  %.PnP Informatio
  // 00000070: 6e09 0200 0901 0209 0201 0905 ac09 0202  n...............
  // 00000070: 6e
  // 00000070:   09 0200                                0x0200 Attr ID: Specification ID 
  // 00000070:           0901 02                        UINT16: 0x0102
  // 00000070:                  09 0201 0905 ac         0x0201 Attr ID: Vendor ID 0x05AC
  // 00000070:                                 09 0202  0x0202 Attr ID: Product ID 
  // 00000080: 0912 9709 0203 0907 0009 0204 2801 0902  ............(...
  // 00000080: 0912 97                                  Prod. ID 0x1297
  // 00000080:        09 0203 0907 00                   0x203 Attr ID: Device Version 0x0700 (7.0)
  // 00000080:                       09 0204 2801       0x0204 Attr ID: Primary Record Boolean (0x28) True (0x01)
  // 00000080:                                    0902  0x0205 Attr ID: Vendor ID Source 
  // 00000090: 0509 0002 00                             UINT16: 0x0002 (USB Implementer's Forum)
  // 00000090:           00                             Continuation State: 0x00 

  0x00, 0x8c, 0x00, 0x89, 0x35, 0x87, 0x35, 0x85, 0x09,
  0x00, 0x00, 0x0a, 0x4f, 0x49, 0x12, 0x00, 0x09, 0x00, 0x01, 0x35, 0x03, 0x19, 0x12, 0x00, 0x09,
  0x00, 0x02, 0x0a, 0x00, 0x00, 0x00, 0x00, 0x09, 0x00, 0x05, 0x35, 0x03, 0x19, 0x10, 0x02, 0x09,
  0x00, 0x06, 0x35, 0x24, 0x09, 0x65, 0x6e, 0x09, 0x00, 0x6a, 0x09, 0x01, 0x00, 0x09, 0x66, 0x72,
  0x09, 0x00, 0x6a, 0x09, 0x01, 0x10, 0x09, 0x64, 0x65, 0x09, 0x00, 0x6a, 0x09, 0x01, 0x20, 0x09,
  0x6a, 0x61, 0x09, 0x00, 0x6a, 0x09, 0x01, 0x30, 0x09, 0x00, 0x08, 0x08, 0xff, 0x09, 0x01, 0x01,
  0x25, 0x0f, 0x50, 0x6e, 0x50, 0x20, 0x49, 0x6e, 0x66, 0x6f, 0x72, 0x6d, 0x61, 0x74, 0x69, 0x6f,
  0x6e, 0x09, 0x02, 0x00, 0x09, 0x01, 0x02, 0x09, 0x02, 0x01, 0x09, 0x05, 0xac, 0x09, 0x02, 0x02,
  0x09, 0x12, 0x97, 0x09, 0x02, 0x03, 0x09, 0x07, 0x00, 0x09, 0x02, 0x04, 0x28, 0x01, 0x09, 0x02,
  0x05, 0x09, 0x00, 0x02, 0x00
};

///////////////////////////////////////////////////////////////////////////
//
// SDP Search Response Templates
// See Core Specification Version 4, Vol. 3 Part B.3.1
// Type  Desc Sizes     Description
// 0     0              Nil, the null type
// 1     0, 1, 2, 3, 4  Unsigned Integer
// 2     0, 1, 2, 3, 4  Signed twos-complement integer
// 3     1, 2, 4        UUID, a universally unique identifier
// 4     5, 6, 7        Text string
// 5     0              Boolean
// 6     5, 6, 7        Data element sequence, a data element whose data field is a sequence of data elements
// 7     5, 6, 7        Data element alternative, data element whose data field is a sequence of data elements from which one data element is to be selected.
// 8     5, 6, 7        URL, a uniform resource locator
// 9-31  Reserved
//
// Size descriptor
// 0     0       1 byte, unless type is nil, in which case datasize is 0
// 1     0       2 bytes
// 2     0       4 bytes
// 3     0       8 bytes
// 4     0       16 bytes
// 5     8       Data size in additional 8 bits
// 6     16      Data size in additional 16
// 7     32      Data size in additional 32 bits
//
// This response must have the structure
//
// - AttributeListsByteCount 
// - AttributeLists
// - ContinuationState
//
// The AttributeLists is a data element sequence
// where each element in turn is a data element sequence
// representing an attribute list.
//
// Each attribute list contains attribute IDs and attribute values
// from one service record.
//
// The first element in each attribute list contains the attribute ID
// of the first attribute to be returned for that service record.
// The second element in each attribute list contains the corresponding
// attribute value.
//
// Successive pairs of elements in each attribute list contain
// additional attribute ID and value pairs. Only attributes that have non-null
// values within the service record and whose attribute IDs were specified
// in the SDP_ServiceSearchAttributeRequest are contained in the AttributeLists.
//
// Neither an attribute ID nor attribute value is placed in AttributeLists
// for attributes in the service record that have no value.
//
// Within each attribute list, the attributes are listed in ascending order of 
// attribute ID value. 
//

const uint8_t SDP_search_attr_response_pbr[] PROGMEM = {/*{{{*/
  SDP_AttributeList(0, // Incomplete DE
    SDP_AttributeList(0, // Incomplete DE
      SDP_ServiceRecordHandle(0x00000001),
      SDP_ServiceClassIDList(0, // Incomplete DE
        // - Service Class - SerialPort UUID - 0x1101
        SDP_ServiceClass(0x1101),
        SDP_CloseContainer()
      ),
      SDP_ServiceRecordState(0x00000000),
      SDP_ServiceID(0x1101),
      SDP_ProtocolDescriptorList(0, // Incomplete DE
        // - Protocol #0: UUID - L2CAP [ 0x0100 ]
        SDP_Protocol(0x0100,0),
        // - Protocol #1: UUID - RFCOMM [0x0003] + ProtocolSpecificParameter #0: Uint8 - Server Channel # 1
        SDP_Protocol(0x0003,2,SDP_UNSIGNED_INT8(0x03)),
        SDP_CloseContainer()
      ),
      SDP_BrowseGroupList(0, // Incomplete DE
        SDP_BrowseGroup(0x1002),
        SDP_CloseContainer()
      ),
      SDP_LanguageBaseAttributeIDList(0, // Incomplete DE
        SDP_LanguageBaseAttributeTuple(0x656e,0x006a,0x0100),
        SDP_CloseContainer()
      ),
      SDP_ServiceAvailability(0xFF), // Full availability
      SDP_BluetoothProfileDescriptorList(0, // Incomplete DE
        // - Profile ID #0: UUID - SerialPortProfile 0x1101, version 0x0201 (2.1)
        SDP_BluetoothProfileElementInt16(0x1101, 0x0201),
        //SDP_BluetoothProfileElementInt16(0x0200, 0x0102), // SpecificationID
        //SDP_BluetoothProfileElementInt16(0x0201, ARDUINO_DUE_VENDOR_ID), // VendorID 
        //SDP_BluetoothProfileElementInt16(0x0202, ARDUINO_DUE_PRODUCT_ID), // ProductID 
        //SDP_BluetoothProfileElementInt16(0x0203, 0x0100), // DeviceVersion
        //SDP_BluetoothProfileElementBoolean(0x0204, 0x01), // PrimaryRecord
        //SDP_BluetoothProfileElementInt16(0x0205, 0x0002), // VendorIDSource 
        SDP_CloseContainer()
      ),
      // FIXME: Including this string, but make sure to include a LanguageBaseAttributeIDList  
      SDP_ServiceName(0x0100, 4, 'C','O','M','1'),
      // SDP_ServiceDescription(0x0100, 11, 'S','e','r','i','a','l',' ','P','o','r','t'),
      SDP_CloseContainer()
    ),
    SDP_CloseContainer()
  ),
  // ContinuationState: Terminator 0x00
  0x00
};/*}}}*/

const uint8_t SDP_service_not_supported[] PROGMEM = {/*{{{*/
  SDP_AttributeList(0, // Incomplete DE
    SDP_CloseContainer()
  ),
  // ContinuationState: Terminator 0x00
  0x00
};/*}}}*/

///////////////////////////////////////////////////////////////////////////

const char * incoming_conn_type[] = { "SCO", "ACL", "eSCO", "Reserved" };

extern bool uhc_get_std_device_descriptor(usb_dev_desc_t *);

BTD::BTD(USBHost *p, const char* name, const char* pin) :
  pUsb(p), // Pointer to USB class instance - mandatory
  hci_last_invoked(0),
  bAddress(0), // Device address - mandatory
  bNumConfigurations(0), // Available configurations
  maxPktSize(0),
  bNumEP(1), // If config descriptor needs to be parsed
  qNextPollTime(0), // Reset NextPollTime
  pollInterval(BTD_ENDPOINT_POLLING_INTERVAL), // FIXME: Originally zero
  bPollEnable(false), // Don't start polling before dongle is connected
  bSspCapable(false), // Attached BTD does not support Secure Simple Pairing
  bEirCapable(false),
  bConnected(false),
  hci_handle(0)
{/*{{{*/
  for (uint8_t i = 0; i < BTD_NUMSERVICES; i++)
    btService[i] = NULL;

  clearAllVariables(); // Set all variables, endpoint structs etc. to default values

  setName(name);
  setPin(pin);

  if (pUsb) // Register in USB subsystem
    pUsb->RegisterDeviceClass(this); // Set devConfig[] entry

  btd_instance = this;

}/*}}}*/


void BTD::setName(const char *name) {/*{{{*/
  memset(btdName, 0, sizeof(btdName));
  memmove(btdName, const_cast<char *>(name), min(strlen(name),sizeof(btdName)-1));
  btdName[sizeof(btdName)-1] = 0;
}/*}}}*/

void BTD::setPin(const char *pin) {/*{{{*/
  memset(btdPin, 0, sizeof(btdPin));
  memmove(btdPin, const_cast<char *>(pin), min(strlen(pin),sizeof(btdPin)-1));
  btdPin[sizeof(btdPin)-1] = 0;
}/*}}}*/

uint32_t BTD::ConfigureDevice(uint32_t parent, uint32_t port, uint32_t lowspeed) {/*{{{*/

};/*}}}*/

uint32_t BTD::Init(uint32_t parent, uint32_t port, uint32_t lowspeed) {/*{{{*/
  
  uint32_t rcode;
  uint32_t alloc_status;
  usb_dev_desc_t device_descriptor;
  UsbDevice *p = NULL;

  TRACE_USBHOST(printf("\r\n%lu - BTD::Init()\r\n", GetTickCount());)

  pUsb->ClearAddressPool();

  if ( !bNumConfigurations ) {/*{{{*/

    clearAllVariables(); // Set all variables, endpoint structs etc. to default values

    TRACE_USBHOST(printf("BTD ConfigureDevice\r\n");)

    if (bAddress) { // Check if address has already been assigned to an instance
      printf("Address in use\r\n");
      return USB_ERROR_CLASS_INSTANCE_ALREADY_IN_USE;
    }

    p = pUsb->GetAddressPool().GetUsbDevicePtr(0); // Get pointer to pseudo device with address 0 assigned
    if (!p) {
      printf("Address not found\r\n");
      return USB_ERROR_ADDRESS_NOT_FOUND_IN_POOL;
    }

    if (!p->epinfo) {
      printf("epinfo is null\r\n");
      return USB_ERROR_EPINFO_IS_NULL;
    }

    if ( !uhc_get_std_device_descriptor(&device_descriptor) ) 
      goto FailGetDevDescr;

    bAddress = pUsb->GetAddressPool().AllocAddress(parent, false, port); // Allocate new address according to device class

    if (!bAddress) {
      TRACE_USBHOST(printf("Out of address space\r\n");)
        return USB_ERROR_OUT_OF_ADDRESS_SPACE_IN_POOL;
    }

    maxPktSize         = device_descriptor.bMaxPacketSize0;
    bNumConfigurations = device_descriptor.bNumConfigurations;

    epInfo[ BTD_CONTROL_PIPE ].maxPktSize = maxPktSize; 
    epInfo[ BTD_CONTROL_PIPE ].uotghs_ptype = UOTGHS_HSTPIPCFG_PTYPE_CTRL;
    epInfo[ BTD_CONTROL_PIPE ].bInterval = pollInterval; 

    VID = device_descriptor.idVendor;
    PID = device_descriptor.idProduct;

    TRACE_USBHOST(printf("BTD::ConfigureDevice       Vendor ID: 0x%4.4x\r\n", VID);)
    TRACE_USBHOST(printf("BTD::ConfigureDevice      Product ID: 0x%4.4x\r\n", PID);)
    TRACE_USBHOST(printf("BTD::ConfigureDevice        bAddress: 0x%4.4x\r\n", bAddress);)
    TRACE_USBHOST(printf("BTD::ConfigureDevice  Configurations: 0x%04.4x\r\n", bNumConfigurations);)
    TRACE_USBHOST(printf("BTD::ConfigureDevice Max Packet Size: %u\r\n", maxPktSize);)

  }/*}}}*/

  TRACE_USBHOST(printf("BTD Init, configurations %d, bAddress %llu\r\n", bNumConfigurations, bAddress);)

  // Get pointer to assigned address record
  p = pUsb->GetAddressPool().GetUsbDevicePtr(bAddress);

  if (!p) {
    printf("Address %llu not found\r\n", bAddress);
    return USB_ERROR_ADDRESS_NOT_FOUND_IN_POOL;
  }

  TRACE_USBHOST(printf("bAddress: 0x%x\r\n",bAddress);)

  p->lowspeed = false;

  // Get pointer to assigned address record
  p = pUsb->GetAddressPool().GetUsbDevicePtr(bAddress);

  if (!p) {
    printf("Address not found\r\n");
    return USB_ERROR_ADDRESS_NOT_FOUND_IN_POOL;
  }

  p->lowspeed = lowspeed;

  // Assign epInfo to epinfo pointer - only EP0 is known
  rcode = pUsb->setEpInfoEntry(bAddress, 1, epInfo);

  if (rcode)
    goto FailSetDevTblEntry;

  TRACE_USBHOST(printf("Checking device VID 0x%04.4x PID 0x%04.4x. Configurations: %u; initial endpoint count %u\r\n", VID, PID, bNumConfigurations, bNumEP);)

  bNumEP = BTD_MAX_ENDPOINTS;
  for ( uint8_t pipe = 0 ; pipe < bNumEP ; pipe++ ) {
    uint32_t dev_addr = uhd_get_configured_address(pipe);
    uint32_t dev_ep   = uhd_get_pipe_endpoint_address(pipe);
    uint32_t endpoint;
    // TODO: Get interface descriptor information from ASF call
    if ( dev_addr == 0 ) {
      continue;
    }
    else {
      switch (dev_ep) {
        case 0x81: endpoint = BTD_EVENT_PIPE; break;
        case 0x82: endpoint = BTD_DATAIN_PIPE; break;
        case 0x02: endpoint = BTD_DATAOUT_PIPE; break;
        default: continue;
      }
    }
    TRACE_USBHOST(printf(" ~+ EP%u:  Address %llu  EP 0x%02.2X\r\n", endpoint, dev_addr, dev_ep);)
    epInfo[endpoint].deviceEpNum = dev_ep;
    epInfo[endpoint].hostPipeNum = pipe;
    epInfo[endpoint].iIoErrorCount = 0;
  }

  for (uint8_t endpoint = 0 ; endpoint < bNumEP ; endpoint++ ) {
    epInfo[endpoint].hostPipeNum = endpoint;
    TRACE_USBHOST(printf("- BTD::Init : OK EP #%u <- %u [0x%02.2x] -> Address %u\r\n", 
      endpoint, 
      epInfo[endpoint].hostPipeNum, 
      epInfo[endpoint].deviceEpNum,
      uhd_get_configured_address(epInfo[endpoint].hostPipeNum)
    );)
  }

  TRACE_USBHOST(pUsb->GetAddressPool().DumpPool();)

  printf("Bluetooth Dongle LM Configuration Done.\r\n");

  disconnect();

  hci_counter         = 0;
  hci_state           = HCI_INIT_STATE;
  connectionWait      = false;
  bPollEnable         = true;

  TRACE_USBHOST(printf("Polling Enabled.\r\n");)

  return 0; // Successful configuration

  /* diagnostic messages */
FailGetDevDescr:
  printf("Failed to get device descriptor. rcode = 0x%04.4x\r\n", rcode);
  goto Fail;

FailSetDevTblEntry:
  printf("Failed to set device table entry. rcode = 0x%04.4x\r\n", rcode);
  goto Fail;

FailGetConfDescr:
  printf("Failed to get device configuration descriptor. rcode = 0x%04.4x\r\n", rcode);
  goto Fail;

FailSetConfDescr:
  printf("Failed to set device configuration descriptor. rcode = 0x%04.4x\r\n", rcode);
  goto Fail;

FailUnknownDevice:
  printf("Unknown device VID 0x%04.4x PID %04.4x\r\n", VID, PID);
  rcode = USB_DEV_CONFIG_ERROR_DEVICE_NOT_SUPPORTED;
Fail:
  printf("BTD Init Failed, error code: %x\r\n", rcode);
  Release();
  return rcode;
}/*}}}*/

void BTD::resetBtServices() {/*{{{*/
  for (uint8_t i = 0; i < BTD_NUMSERVICES; i++) {
    if (!(NULL == btService[i])) {
      TRACE_USBHOST(printf("Resetting service #%u\r\n", i);)
      btService[i]->Reset(); // Reset all Bluetooth services
    }
  }
}/*}}}*/

void BTD::clearAllVariables() {/*{{{*/
  irqflags_t flags = cpu_irq_save();
  transfer_queue_flush();
  resetBtServices();
  for (uint8_t i = 0; i < BTD_MAX_ENDPOINTS; i++) {
    epInfo[i].deviceEpNum     = 0;
    epInfo[i].maxPktSize      = (i) ? 0 : 8;
    epInfo[i].epAttribs       = 0;
    epInfo[i].bmNakPower      = (i) ? USB_NAK_NOWAIT : USB_NAK_MAX_POWER;
    epInfo[i].bDescriptorType = 0;
    epInfo[i].uotghs_ptype    = 0;
  }
  pollInterval       = BTD_ENDPOINT_POLLING_INTERVAL;
  bPollEnable        = false; // Don't start polling before dongle is connected
  bAddress           = 0;     // Clear device address
  hci_last_invoked   = NULL;
  bNumConfigurations = 0;     // Available configurations
  maxPktSize         = 0;
  bNumEP             = 1;     // Must have to be reset to 1
  qNextPollTime      = 0;     // Reset next poll time
  hci_counter        = 0;
  cpu_irq_restore(flags);
}/*}}}*/

void BTD::PrintEndpointDescriptor(const USB_ENDPOINT_DESCRIPTOR* ep_ptr) {/*{{{*/
  static const char * typeList[] = {
    "UNSPEC",
    "DEVICE",
    "CONFIGURATION",
    "STRING",
    "INTERFACE",
    "ENDPOINT",
    "DEVICE_QUALIFIER",
    "OTHER_SPEED",
    "INTERFACE_POWER",
    "OTG"
  };
  printf("Endpoint descriptor:\r\n");
  printf("\t    Length:\t0x%02.2x\r\n",ep_ptr->bLength);
  printf("\t      Type:\t0x%02.2x USB_DESCRIPTOR_%s\r\n",ep_ptr->bDescriptorType, (ep_ptr->bDescriptorType < 10 ? typeList[ep_ptr->bDescriptorType] : typeList[0]));
  printf("\t EndptAddr:\t0x%02.2x (%s)\r\n",ep_ptr->bEndpointAddress,ep_ptr->bEndpointAddress & 0x80 ? "In" : "Out");
  printf("\tAttributes:\t0x%02.2x\r\n",ep_ptr->bmAttributes);
  printf("\tMaxPktSize:\t0x%02.2x\r\n",ep_ptr->wMaxPacketSize);
  printf("\tPoll Intrv:\t0x%02.2x\r\n",ep_ptr->bInterval);
}/*}}}*/

/* Performs a cleanup after failed Init() attempt */

uint32_t BTD::Release() {/*{{{*/
  clearAllVariables(); // Set all variables, endpoint structs etc. to default values
  pUsb->GetAddressPool().FreeAddress(bAddress);
  return 0;
}/*}}}*/

uint32_t BTD::Poll() {/*{{{*/

  // Non-IHC context (invoked by Usb::Task())
  if (!bPollEnable) return 0;
  // FIXME: Freeze on error, no further polling allowed.
  if (pUsb->getUsbTaskState() != USB_STATE_RUNNING) return 0;
  if (qNextPollTime > GetTickCount()) return 0;

  // Poll the HCI event pipe
  HCI_event_task();

  // Poll the L2CAP pipe (bulk-IN 0x82), dispatch protocol commands.
  ACL_event_task();

  consume_queue();

  qNextPollTime = GetTickCount() + pollInterval; // Set new poll time

  return 0;
}/*}}}*/

/*
 * \brief Dispatch a queued command from the polling loop 
 */
bool BTD::consume_queue()
{/*{{{*/

  /*
   * @return false - No command dispatched, command queue empty, or failed to poll bulk-in endpoint
   *         true  - Nonempty command queue, setup of read on bulk-in endpoint succeeded.
   */
  // IHC
  
  static uhd_l2cap_queue_t e;
  uint8_t service_class_id[2];
  uint8_t device_class_id[2];
  uint8_t index;

  sdp_pdu_header_t * sh;
  sdp_data_element_t sdp;
  uint8_t sp_counter = 0;
  uint8_t * des_token_pos;

  if ( transfer_queue_empty() ) {
    return true;
  }

  switch ( e.type ) {
    case L2CAP_QUEUE_INFORMATION_RESPONSE:
    case L2CAP_QUEUE_INFORMATION_REQUEST:
    case L2CAP_SERVICE_NOT_SUPPORTED:
    case L2CAP_SDP_SERVICE_ATTRIBUTE_RESPONSE_PBR:
    case L2CAP_CONNECTION_RESPONSE:
    case L2CAP_DISCONNECT_REQUEST:
    case L2CAP_DISCONNECT_RESP:
    case L2CAP_CONFIG_REQUEST:
    case L2CAP_CONFIG_RESPONSE:
    case RFCOMM_PROTOCOL_MESSAGE:
    case RFCOMM_PROTOCOL_CREDIT:
      if (pipe_busy(bAddress, epInfo[BTD_DATAOUT_PIPE].deviceEpNum)) return false;
      break;
  }


  if ( tq_next(&e) )
  {/*{{{*/

    // Dispatch queued command 

    switch ( e.type ) {
      case L2CAP_SERVICE_NOT_SUPPORTED:
      case L2CAP_SDP_SERVICE_ATTRIBUTE_RESPONSE_PBR:
        service_class_id[0] = e.t.sdp.service_class_id & 0xFF;
        service_class_id[1] = (e.t.sdp.service_class_id >> 8) & 0xFF;
        break;
      case L2CAP_CONNECTION_RESPONSE:
      case L2CAP_DISCONNECT_REQUEST:
      case L2CAP_DISCONNECT_RESP:
        service_class_id[0] = e.t.conn_resp.scid & 0xFF;
        service_class_id[1] = (e.t.conn_resp.scid >> 8) & 0xFF;
        device_class_id[0]  = e.t.conn_resp.dcid & 0xFF;
        device_class_id[1]  = (e.t.conn_resp.dcid >> 8) & 0xFF;
        break;
      case L2CAP_CONFIG_REQUEST:
      case L2CAP_CONFIG_RESPONSE:
        service_class_id[0] = e.t.conf_req.scid & 0xFF;
        service_class_id[1] = (e.t.conf_req.scid >> 8) & 0xFF;
        break;
      case RFCOMM_PROTOCOL_MESSAGE:
      case RFCOMM_PROTOCOL_CREDIT:
        service_class_id[0] = e.t.rfcomm.scid % 0x100;
        service_class_id[1] = e.t.rfcomm.scid / 0x100;
        break;
    }

    switch ( e.type ) {

      // Only the PDU ID and Transaction ID have been copied into
      // the output buffer:
      // - SDP_SERVICE_SEARCH_ATTRIBUTE_RESPONSE_PDU
      // - Transaction ID
      //
      // These items will be computed based on the content of 
      // SDP_search_attr_response_pbr:
      // - TotalServiceRecordCount
      // - CurrentServiceRecordCount
      // - ServiceRecordHandleList
      // - ContinuationState

      // Parse the SDP message into position 7 of the output buffer, and 
      // store these 2-byte values in the positions indicated:
      // 0x0003 - 2 bytes, Total parameter length
      // 0x0005 - 2 bytes, Attribute List byte count

      case L2CAP_SERVICE_NOT_SUPPORTED:

        memmove(l2capoutbuf, e.l2capoutbuf, e.message_length);
        index = prepare_sdp_pdu_message(
          &(l2capoutbuf[7]),
          &(SDP_service_not_supported[0]),
          sizeof(SDP_service_not_supported)
        );

        break;

      case L2CAP_SDP_SERVICE_ATTRIBUTE_RESPONSE_PBR:

        memmove(l2capoutbuf, e.l2capoutbuf, e.message_length);
        index = prepare_sdp_pdu_message(
          &(l2capoutbuf[7]),
          &(SDP_search_attr_response_pbr[0]),
          sizeof(SDP_search_attr_response_pbr)
        );

        break;
    }

    switch ( e.type ) {

      case L2CAP_SERVICE_NOT_SUPPORTED:
      case L2CAP_SDP_SERVICE_ATTRIBUTE_RESPONSE_PBR:

        // Response length sans Continuation marker
        *((uint16_t *)&(l2capoutbuf[5])) = byteflip((uint16_t)index - 1);
        // Response length including Continuation marker
        *((uint16_t *)&(l2capoutbuf[3])) = byteflip((uint16_t)(index + 2));

        e.message_length += index;

        TRACE_USBHOST(printf("%lu Issue %s, %u bytes\r\n", GetTickCount(), l2cap_message_types[e.type], e.message_length);)

        L2CAP_Command(
          e.handle,
          l2capoutbuf,
          e.message_length,
          service_class_id[0], //  0x01,
          service_class_id[1], //  0x00,
          NULL, // Force default callback
          e.queue_index + 1
        );

        break;

      case L2CAP_QUEUE_INFORMATION_RESPONSE:
      case L2CAP_QUEUE_INFORMATION_REQUEST:
      case L2CAP_CONNECTION_RESPONSE:
      case L2CAP_DISCONNECT_REQUEST:
      case L2CAP_DISCONNECT_RESP:
      case L2CAP_CONFIG_REQUEST:
      case L2CAP_CONFIG_RESPONSE:

        TRACE_USBHOST(printf("%lu Issue %s, %u bytes\r\n", GetTickCount(), l2cap_message_types[e.type], e.message_length);)
        memmove(l2capoutbuf, e.l2capoutbuf, e.message_length);
        L2CAP_Command(
          e.handle,
          l2capoutbuf,
          e.message_length,
          0x01, 0x00, 
          NULL,
          e.queue_index + 1
        );

        break;

      /* Device type-specific messages */
      case RFCOMM_PROTOCOL_MESSAGE:
        TRACE_USBHOST(printf("%lu Issue RFCOMM_PROTOCOL_MESSAGE\r\n", GetTickCount());)
        l2capoutbuf[0] = e.t.rfcomm.address;
        l2capoutbuf[1] = e.t.rfcomm.control;
        l2capoutbuf[2] = e.t.rfcomm.length << 1 | 0x01;
        for (index = 0; index < e.t.rfcomm.length; index++)
          l2capoutbuf[index + 3] = e.l2capoutbuf[index];
        l2capoutbuf[index + 3] = calcFcs(l2capoutbuf);

        L2CAP_Command(
          e.handle,
          l2capoutbuf,
          index + 4, // Final length includes FCS byte
          service_class_id[0],
          service_class_id[1],
          RFCOMM_PROTOCOL_MESSAGE_handler,
          e.queue_index + 1
        );

        break;

      case RFCOMM_PROTOCOL_CREDIT:
        TRACE_USBHOST(printf("%lu Issue RFCOMM_PROTOCOL_CREDIT\r\n", GetTickCount());)
        l2capoutbuf[0] = e.t.rfcomm.address;
        l2capoutbuf[1] = e.t.rfcomm.control;
        l2capoutbuf[2] = e.t.rfcomm.length << 1 | 0x01;
        l2capoutbuf[3] = e.l2capoutbuf[0];
        l2capoutbuf[4] = calcFcs(l2capoutbuf);
        
        L2CAP_Command(
          e.handle,
          l2capoutbuf,
          5,
          service_class_id[0],
          service_class_id[1],
          RFCOMM_PROTOCOL_CREDIT_handler,
          e.queue_index + 1
        );
        break;
    }
    return true; // Poll the L2CAP pipe (bulk-IN 0x82), dispatch protocol commands.
  }/*}}}*/

  return false;
}/*}}}*/

uint8_t BTD::get_service_index(uint8_t callback_data) {/*{{{*/
  if ( callback_data > 0 ) {
    callback_data--;
    return tq_at(callback_data).bt_service_index; 
  }
  // If this value is received, do not use the queue item.
  return BTD_NUMSERVICES;
}/*}}}*/

uint32_t BTD::get_instance_payload(uint8_t queue_index) {
  if ( queue_index && (queue_index < BTD_NUMSERVICES) ) {
    uint32_t payload = L2CAP_QUEUE_NOOP;
    switch ( tq_at(queue_index).type  ) {
      case L2CAP_CONNECTION_RESPONSE:
        payload = tq_at(queue_index).t.conn_resp.payload;
        break;
      case L2CAP_SERVICE_NOT_SUPPORTED:
        payload = tq_at(queue_index).t.sdp.payload;
        break;
      case L2CAP_CONFIG_REQUEST:
      case L2CAP_CONFIG_RESPONSE:
        payload = tq_at(queue_index).t.conf_req.payload;
        break;
      case L2CAP_DISCONNECT_REQUEST:
      case L2CAP_DISCONNECT_RESP:
        payload = tq_at(queue_index).t.disc.payload;
        break;
      default:
        printf("? Unhandled L2CAP_Command type %u\r\n",
          tq_at(queue_index).type
        ); 
    }
    return payload;
  }
  // If the index exceeds array bounds, perform no-op 
  return L2CAP_QUEUE_NOOP;
}

void BTD::invoke_callback_state_update(usb_add_t add,
    usb_ep_t ep,
    uhd_trans_status_t status,
    iram_size_t nb_transferred,
    uint8_t callback_data) {/*{{{*/

  if ( callback_data == 0 ) {
    printf("~ No state change. Transferred %u bytes, status 0x%02.2X\r\n",
      nb_transferred, status
    );
  }
  else {
    uint8_t bt_service_index = get_service_index(callback_data);
    uint8_t state_change_flag = get_instance_payload(bt_service_index);

    if ( bt_service_index < BTD_NUMSERVICES )
      btService[bt_service_index]->StateChange(state_change_flag);
  }

}/*}}}*/

void BTD::ACL_state_update(uhd_trans_status_t status)
{/*{{{*/

  // We are in IHC (Interrupt-handling context)
  // for dealing with messages 
  // on the Bluetooth device bulk-IN endpoint
  // It is possible to dispatch single messages on these pipes:
  //   bulk-OUT
  //   control-OUT
  //   or event (interrupt) IN

  // printf("# ACL_state_update(0x%02.2X)\r\n",status);
  int e = 0;
  for (uint8_t bt_service_index = 0; bt_service_index < BTD_NUMSERVICES; bt_service_index++) {
    if (!(NULL == btService[bt_service_index])) {
      btService[bt_service_index]->Run(l2capinbuf);
      e++;
    }
  }
  // At this point, zero or more Bluetooth services
  // have enqueued protocol command messages, 
  // which need to be dispatched one at a time,
  // and handled outside this context.
  pUsb->setErrFlag(e == 0);

}/*}}}*/

void ACL_event_task_(usb_add_t add,
    usb_ep_t ep,
    uhd_trans_status_t status,
    iram_size_t nb_transferred,
    uint8_t callback_data) 
{/*{{{*/
  // Bulk-IN endpoint IHC, job complete, dispatch on any pipe possible.
  btd_instance->ACL_state_update(status);
}/*}}}*/

bool BTD::ACL_event_task()
{/*{{{*/

  // Non-IHC
  return uhd_ep_run(
    bAddress,
    epInfo[BTD_DATAIN_PIPE].deviceEpNum,
    true, 
    this->l2capinbuf,
    L2CAP_IN_BUFFERSIZE, // Maximum size is 64 bytes
    // A timeout occurs after n milliseconds of waiting,
    // following which the ACL_event_task_() method is invoked 
    20,
    // This method is invoked from an interrupt context
    ACL_event_task_,
    0
  );

}/*}}}*/

void HCI_event_task_(usb_add_t add,
    usb_ep_t ep,
    uhd_trans_status_t status,
    iram_size_t nb_transferred,
    uint8_t callback_data) 
{/*{{{*/
  btd_instance->HCI_state_update(status);
}/*}}}*/

bool BTD::HCI_event_task(uhd_trans_status_t status)
{/*{{{*/

  // Non-blocking implementation of HCI event pipe polling.
  // The Bluetooth device event pipe (interrupt-in) is read by
  // queueing a read request on that pipe from the ASF framework.
  // When that call concludes (or hits timeout), the HCI state
  // machine is updated in an interrupt-handler context.
  // This means that the handler:
  // 1) Cannot issue another read of the event pipe.
  // 2) Must test whether the device bulk I/O pipes are not currently busy,
  //    else transmission should be deferred on those pipes.
  // 3) Can issue control pipe messages.
  return uhd_ep_run(
    bAddress,
    epInfo[BTD_EVENT_PIPE].deviceEpNum,
    true, 
    this->hcibuf,
    HCI_BUFFERSIZE,
    20,
    HCI_event_task_,
    0
  );

}/*}}}*/

void BTD::HCI_state_update(uhd_trans_status_t status) {/*{{{*/
  // Note: Spec 4.0 Vol. 2, 5.4.4 Interrupt-IN packet limit: 255 bytes.

  uint8_t param_length = hcibuf[1];
  bool dump_hcibuf     = false;

  ticks = GetTickCount();

  switch (hcibuf[0]) //switch on event type
  {
    case EV_COMMAND_COMPLETE:
      switch (hci_state) {/*{{{*/
        case HCI_EXTENDED_INQUIRY_RESPONSE_SENT_STATE:
          hci_event_flag |= HCI_FLAG_CMD_COMPLETE;
          break;
        case HCI_READ_LOCAL_SUPPORTED_FEATURES:
          // See 2.C.3.3, Feature Mask Definition
          hci_event_flag |= HCI_FLAG_CMD_COMPLETE;
          memmove(local_features_bitmask, &(hcibuf[6]), 8);
#ifdef ALLOW_SSP
          bSspCapable = hcibuf[6 + 6] & 0x08; // Secure Simple Pairing flag, see CV4, Vol. 2.C.4.2.7 
#ifdef DEBUG_HCI_EVENTS
          printf("%lu - Feature bit dump. SSP support %s", ticks, bSspCapable ? "YES" : "NO");
#endif
#else
          bSspCapable = false;
#ifdef DEBUG_HCI_EVENTS
          printf("%lu - SSP disallowed. SSP support %s", ticks, bSspCapable ? "YES" : "NO");
#endif
#endif
          bEirCapable = (hcibuf[6 + 6] & 0x01) == 0x01; // Extended Inquiry Response capability 
#ifdef DEBUG_HCI_EVENTS
          printf(" EIR support %s\r\n", bEirCapable ? "YES" : "NO");
#endif
          break;
        case HCI_READ_LOCAL_OOB_DATA_STATE:
          hci_event_flag |= HCI_FLAG_CMD_COMPLETE;
          memset(oob_pairing_hash, 0, sizeof(oob_pairing_hash));
          memset(oob_pairing_randomizer, 0, sizeof(oob_pairing_randomizer));
          if ( !hcibuf[5] ) {
            memmove(oob_pairing_hash, &(hcibuf[6]), 16);
            memmove(oob_pairing_randomizer, &(hcibuf[22]), 16);
#ifdef DEBUG_HCI_EVENTS
            printf("%lu - Obtained SSP pairing precursor data\r\n", ticks);
            printf("%lu - Pairing Hash: ", ticks);
            for ( uint8_t h = 0 ; h < 16 ; h++ ) printf("0x%02.2X ", oob_pairing_hash[h]); printf("\r\n"); 
            printf("%lu -   Randomizer: ", ticks);
            for ( uint8_t h = 0 ; h < 16 ; h++ ) printf("0x%02.2X ", oob_pairing_randomizer[h]); printf("\r\n"); 
#endif
          }
          break;
        case HCI_SET_HOST_BUFFER_SIZE:
#ifdef DEBUG_HCI_EVENTS
          printf("%lu - Set host buffer size\r\n", ticks);
#endif
          hci_event_flag |= HCI_FLAG_CMD_COMPLETE;
          break;
        case HCI_READ_HOST_BUFFER_SIZE:
          hci_event_flag |= HCI_FLAG_CMD_COMPLETE;
#ifdef DEBUG_HCI_EVENTS
          // 3, 4: OCF, OGF
          // 5   : Status
          printf("%lu   HCI ACL Data Packet Length %u (0x%02.2X%02.2X)\r\n", ticks, (uint16_t)(hcibuf[6] | (hcibuf[7] << 8)), hcibuf[7], hcibuf[6] );
          printf("%lu   HCI Synchronous Data Packet Length %u (0x%02.2X)\r\n", ticks, hcibuf[8], hcibuf[8] );
          printf("%lu   Total ACL packets %u (0x%02.2X%02.2X)\r\n", ticks, (uint16_t)(hcibuf[9] | (hcibuf[10] << 8)), hcibuf[10], hcibuf[9] );
          printf("%lu   Total Synchronous packets %u (0x%02.2X%02.2X)\r\n", ticks, (uint16_t)(hcibuf[11] | (hcibuf[12] << 8)), hcibuf[12], hcibuf[11] );
#endif
          break;
        default:
          if (!hcibuf[5]) { // Check if command succeeded
            hci_event_flag |= HCI_FLAG_CMD_COMPLETE; // set command complete flag
            if ((hcibuf[3] == 0x01) && (hcibuf[4] == 0x10)) { // parameters from read local version information
              hci_version = hcibuf[6]; // Used to check if it supports 2.0+EDR - see http://www.bluetooth.org/Technical/AssignedNumbers/hci.htm
              hci_event_flag |= HCI_FLAG_READ_VERSION;
            } else if ((hcibuf[3] == 0x09) && (hcibuf[4] == 0x10)) { // parameters from read local bluetooth address
              for (uint8_t i = 0; i < 6; i++)
                my_bdaddr[i] = hcibuf[6 + i];
              hci_event_flag |= HCI_FLAG_READ_BDADDR;
            }
          }
      }/*}}}*/
      break;

    case EV_COMMAND_STATUS:
      if (hci_state == HCI_ACCEPT_CONNECTION_STATE ) {/*{{{*/
        printf("%lu - Accept_Connection_Request transmitted. ", ticks );
        hci_dump_event();
      }/*}}}*/
      else if (hci_state == HCI_FLOW_CONTROL_ENABLE) {/*{{{*/
        if ( hcibuf[2] == 0x00 ) {
          printf("%lu - HCI Flow Control Setting OK. HCI PTL %u\r\n", ticks, hcibuf[1] );
          printf("%lu -   Host to Ctrlr: %u\r\n", ticks, hcibuf[3] );
          printf("%lu -          Opcode: OCF 0x%02.2X OGF %02.2X\r\n", ticks, hcibuf[4], hcibuf[5] >> 2 );
          printf("%lu -          Retval: 0x%02.2X\r\n", ticks, hcibuf[6] );
        }
        else {
          printf("%lu - HCI Flow Control Setting Chatter.", ticks );
        }
        hci_dump_event();
      }/*}}}*/
      else if (hcibuf[2]) { // show status on serial if not OK
        printf("%lu - HCI Command Failed (previously executed: 0x%08.8X)", ticks, hci_last_invoked);
        dump_hcibuf = true;
      }
      break;

    case EV_HARDWARE_ERROR:
      printf("! Hardware Error Event evlen 0x%04.4x", hcibuf[1]);
      for (uint16_t i = 0; i < HCI_BUFFERSIZE; i++) {
        if (!(i % 16)) printf("\r\n%lu ", GetTickCount());
        if (!(i % 8)) printf(" ");
        printf(" %02.2x", hcibuf[i]);
      }
      printf("\r\n");

      break;

    case EV_INQUIRY_COMPLETE:
      if (inquiry_counter >= 5) {
        inquiry_counter = 0;
        TRACE_USBHOST(printf("Inquiry complete; enter HCI_SCANNING_STATE\r\n\r\n");)
          hci_state = HCI_SCANNING_STATE;
      }
      inquiry_counter++;
      break;

    case EV_INQUIRY_RESULT:
      if (hcibuf[2]) { // Check that there is more than zero responses
#ifdef EXTRADEBUG
        TRACE_USBHOST(printf("Number of responses: %d\r\n",hcibuf[2]);)
#endif
        for (uint8_t i = 0; i < hcibuf[2]; i++) {
          hci_event_flag |= HCI_FLAG_NONEMPTY_INQUIRY_RESULT;
#ifdef EXTRADEBUG
          TRACE_USBHOST(printf("Class of device: 0x%02.2x %02.2x %02.2x\r\n",hcibuf[6 + 8 * hcibuf[2] + 3 * i], hcibuf[5 + 8 * hcibuf[2] + 3 * i], hcibuf[4 + 8 * hcibuf[2] + 3 * i] );)
#endif
        }
      }
      break;

    case EV_USER_PASSKEY_NOTIFICATION:
      printf("! User Passkey Notification from host ");
      for ( uint32_t b = 0 ; b < 6 ; b++ ) {
        printf("%02.2X ", hcibuf[10-b]);
      }
      printf("\r\n");
      break;

    case EV_CONNECT_COMPLETE:
      if (!hcibuf[2]) { // check if connected OK
        hci_handle = hcibuf[3] | ((hcibuf[4] & 0x0F) << 8); // store the handle for the ACL connection
        hci_event_flag |= HCI_FLAG_CONN_COMPLETE; // set connection complete flag
        hci_state = HCI_HANDLE_RECEIVED;
        // hci_state = HCI_CONNECTED_STATE;
        bConnected = true;

        printf("HCI Handle 0x%04.4X for %s connection.\r\n",hci_handle, incoming_conn_type[hcibuf[11] < 0x03 ? hcibuf[11] & 0x3 : 3] );
        printf("Remote Bluetooth Address: ");
        for ( uint32_t b = 0 ; b < 6 ; b++ ) {
          printf("%02.2X ", hcibuf[10-b]);
        }
        printf("\r\n");
        printf("Stored Bluetooth Address: ");
        for ( uint32_t b = 0 ; b < 6 ; b++ ) {
          printf("%02.2X ", disc_bdaddr[5-b]);
        }
        printf("\r\n");
        dump_hcibuf = true;
      }
      else {
        printf("\r\n\n!! Connection Failed\r\n");
        hci_dump_event();
        hci_event_flag = 0;
        hci_counter = 0;
        hci_write_scan_disable();
        hci_state = HCI_INIT_STATE;
      }
      break;

    case EV_DISCONNECT_COMPLETE:
      if (!hcibuf[2]) { // check if disconnected OK
        hci_event_flag |= HCI_FLAG_DISCONN_COMPLETE; // set disconnect command complete flag
        hci_event_flag &= ~HCI_FLAG_CONN_COMPLETE; // clear connection complete flag
        bConnected = false;
        hci_handle = 0x0000;
      }
      break;

    case EV_REMOTE_NAME_COMPLETE:
      if (!hcibuf[2]) { // check if reading is OK
        for (uint8_t i = 0; i < sizeof(remote_name); i++)
          remote_name[i] = hcibuf[9 + i];
        hci_event_flag |= HCI_FLAG_REMOTE_NAME_COMPLETE;
      }
      break;

    case EV_ENCRYPTION_CHANGE:
      printf("! Enc change.\r\n");
      hci_event_flag |= HCI_FLAG_ENCRYPTION_CHANGE_EVENT;
      break;

    case EV_INCOMING_CONNECT:
      hci_event_flag |= HCI_FLAG_INCOMING_REQUEST;
      hci_state = HCI_CONNECT_IN_STATE;
      memmove(disc_bdaddr, &(hcibuf[2]), 6);
      hci_dump_event();
      printf("Class of device: 0x%02.2x %02.2x %02.2x requesting %s connection.\r\n",hcibuf[10], hcibuf[9], hcibuf[8], incoming_conn_type[hcibuf[11] < 0x03 ? hcibuf[11] & 0x3 : 3] );
      printf("Remote Bluetooth Address: ");
      for ( uint32_t b = 0 ; b < 6 ; b++ ) {
        if ( !( b % 8 ) ) printf("\r\n");
        printf("%02.2X ", disc_bdaddr[5-b]);
      }
      printf("\r\n");
      bConnected = false;
      break;

    case EV_PIN_CODE_REQUEST:
      printf("Bluetooth pin is set to: %s\r\n", &btdPin);
      hci_pin_code_request_reply();
      break;

    case EV_LINK_KEY_REQUEST:
#ifdef KEY_REQUEST_IMPLEMENTED_SAFER
      hci_link_key_request_reply();
#else
      printf("! Received Key Request. No SAFER+ implementation.\r\n");
      hci_link_key_request_negative_reply();
#endif
      break;

    case EV_FLOW_SPECIFICATION_COMPLETE:

      hci_event_flag |= HCI_FLAG_FLOW_SPECIFICATION;
      printf("%lu - HCI Flow Specification Event. Total Length: %u\r\n", ticks, hcibuf[1] );
      printf("%lu -     FSE: %s\r\n", ticks, hcibuf[2] == 0 ? "OK" : "Fail" );
      printf("%lu -  Handle: 0x%02.2X%02.2X", ticks, hcibuf[4], hcibuf[3] );
      dump_hcibuf = true;

      break;


    case EV_LINK_KEY_NOTIFICATION:
      hci_event_flag |= HCI_FLAG_LINK_KEY_NOTIFICATION_EVENT;
      hci_state = HCI_LINK_KEY_NOTIFICATION_EVENT;
      // Store the link key received from the remote peer
      memset( link_key, 0, sizeof(link_key) );
      memmove( link_key, &(hcibuf[8]), 16 );
      link_key_type = hcibuf[24];

      printf("%lu - Received Link Key Notification.", ticks);
      hci_dump_event();
      printf("%lu   - Link Key Type: 0x%02.2X\r\n", ticks, link_key_type);
      printf("%lu   - Link Key: ", ticks);
      for ( uint32_t b = 0 ; b < sizeof(link_key) ; b++ ) {
        if ( !( b % 8 ) ) printf("\r\n");
        printf("%02.2X ", link_key[b]);
      }
      printf("\r\n\n");

      break;

    case EV_AUTHENTICATION_COMPLETE:
      printf("Pairing successful\r\n");
      break;
    case EV_PAGE_SCAN_REP_MODE: // 7.7.31 Page Scan Repetition Mode Change Event
      printf("! Remote Page Scan Repetition mode change\r\n");
      break;
    case EV_MAX_SLOTS_CHANGE: // 7.7.27 Max Slots Change Event 
      printf("! Packet slot length change to %u for handle 0x%02.2X%02.2X\r\n", hcibuf[4], hcibuf[3], hcibuf[2] );
      break;
    case EV_CONNECTION_PACKET_TYPE_CHANGED:
      printf("! Connection Packet Type Changed\r\n");
      dump_hcibuf = true;
      break;
      /* We will just ignore the following events */
    case EV_ROLE_CHANGED:
    case EV_LOOPBACK_COMMAND:
    case EV_CHANGE_CONNECTION_LINK:
    case EV_QOS_SETUP_COMPLETE:
      printf("Unhandled HCI Event: 0x%02.2x, last command 0x%08.8X\r\n",hcibuf[0], hci_last_invoked);
      break;
    case EV_NUM_COMPLETE_PKT:
#ifdef DEBUG_HCI_EVENTS
      printf("! Complete Packets event: 0x%02.2x, last command 0x%08.8X\r\n",hcibuf[0], hci_last_invoked);
#endif
      // hci_dump_completed_packets_event(hcibuf);
      completed_packet_events++;
      // dump_hcibuf = true;
      break;
    case EV_DATA_BUFFER_OVERFLOW:
      printf("! Data Buffer Overflow event: 0x%02.2x, last command 0x%08.8X\r\n",hcibuf[0], hci_last_invoked);
      break;
    case EV_IO_CAPABILITY_REQUEST : // 0x31; send IO capability response
      printf("! EV_IO_CAPABILITY_REQUEST\r\n");
      hci_event_flag |= HCI_FLAG_IO_CAPABILITY_REQUEST; 
      hci_state = HCI_IO_CAPABILITY_REQUESTED_STATE;
      dump_hcibuf = true;
      break;
    case EV_IO_CAPABILITY_REPLY   : // 0x32
      printf("! EV_IO_CAPABILITY_REPLY\r\n");
      break;
    case EV_RETURN_LINK_KEYS:
      printf("~ Stored link key event\r\n");
      break;
    case EV_READ_REMOTE_VERSION_INFORMATION_COMPLETE:
      printf("~ Remote Version Info OK\r\n");
      hci_event_flag |= HCI_FLAG_READ_REMOTE_VERSION_INFORMATION_EVENT;
      break;
    case EV_READ_REMOTE_SUPPORTED_FEATURES_COMPLETE:
      printf("~ Remote Supp Feat Read OK\r\n");
      hci_event_flag |= HCI_FLAG_READ_REMOTE_SUPPORTED_FEATURES_EVENT;
      break;
    case EV_READ_REMOTE_EXTENDED_FEATURES_COMPLETE:
      printf("~ Remote Ext Feat Read OK\r\n");
      hci_event_flag |= HCI_FLAG_READ_REMOTE_EXTENDED_FEATURES_EVENT;
      dump_hcibuf = true;
      break;
    case EV_USER_CONFIRMATION_REQUEST:
      // See 7.7.42 User Confirmation Request Event
      printf("%lu - ! User Confirmation Requested\r\n", ticks);
      hci_state = HCI_USER_CONFIRMATION_REQUEST;
      break;
    case EV_OOB_DATA_REQUEST_EVENT: // Async rq for OOB data, 2.E.7.7.44 
      printf("%lu - ! OOB Data Req\r\n", ticks);
      hci_event_flag |= HCI_FLAG_OOB_DATA_REQUEST_EVENT; 
      hci_state = HCI_OOB_DATA_REQUEST_STATE;
      break;
    case EV_SIMPLE_PAIRING_COMPLETE:
      printf("%lu - ! Simple Pairing complete\r\n", ticks);
      hci_event_flag |= HCI_FLAG_SIMPLE_PAIRING_COMPLETE;
      hci_state = HCI_SIMPLE_PAIRING_COMPLETE;
      break;
    default:
      if (hcibuf[0] != 0x00) {
        printf("%lu %s HCI Event: 0x%02.2x, last command 0x%08.8X\r\n",
          GetTickCount(),
          // See, for example, bcore-sp-009P, Sec. 3.2, "Manufacturer-Specific HCI Commands and Events"
          hcibuf[0] == 0xFF ? "Mfg." : "Unhandled",
          hcibuf[0],
          hci_last_invoked
        );
        dump_hcibuf = true;
      }
      break;
  } // switch

  if (dump_hcibuf) hci_dump_event();

  // Prevent duplicate processing of event.
  hcibuf[0] = 0;
  HCI_task();

}/*}}}*/


void BTD::HCI_task() {/*{{{*/

  // HCI state machine

  ticks = GetTickCount();
  uint32_t transfer_length;

  switch (hci_state)
  {/*{{{*/

    case HCI_INIT_STATE:
      hci_counter++;
      if (hci_counter > RESET_DEFERRAL_LOOPS) {
        // wait until we have looped x times to clear any old events
        hci_reset();
        // >= 2 * 10 millisecond pause after reset (USB 2.0 9.2.6.2)
        qNextPollTime = GetTickCount() + 100;
        hci_counter = 0;
      }
      break;

    case HCI_RESET_STATE:
      // if (hci_cmd_complete) hci_delete_stored_link_key(NULL);
      if (hci_cmd_complete) hci_set_controller_to_host_flow_control();
      break;

    case HCI_DELETE_STORED_LINK_KEYS:
      if (hci_cmd_complete) hci_set_controller_to_host_flow_control();
      break;

    case HCI_C_TO_H_FLOW_CONTROL_ENABLE:
      if (hci_cmd_complete) hci_read_host_buffer_size();
      break;

    case HCI_READ_HOST_BUFFER_SIZE:
      if (hci_cmd_complete) hci_set_host_buffer_size();
      break;

    case HCI_SET_HOST_BUFFER_SIZE:
      if (hci_cmd_complete) hci_read_local_supported_commands();
      break;

    case HCI_READ_LOCAL_SUPPORTED_COMMANDS:
      if (hci_cmd_complete) hci_read_local_supported_features();
      break;

    case HCI_READ_LOCAL_SUPPORTED_FEATURES:
      if (hci_cmd_complete) {
        if (bSspCapable) {
          hci_write_simple_pairing_mode(bSspCapable);
        } 
        else {
          hci_write_class_of_device();
        }
      }
      break;

    case HCI_WRITE_SIMPLE_PAIRING_MODE_STATE:
      if (hci_cmd_complete) hci_read_local_oob_data();
      break;

    case HCI_READ_LOCAL_OOB_DATA_STATE:
      if (hci_cmd_complete) hci_write_class_of_device();
      break;

    case HCI_CLASS_STATE:
      if (hci_cmd_complete || noresp_limit) hci_set_default_event_mask();
      break;

    case HCI_SET_EVTMASK_STATE:
      if (hci_cmd_complete || noresp_limit) hci_clear_event_filter();
      break;

    case HCI_CLEAR_EVENTS_STATE:
      if (hci_cmd_complete) hci_set_default_event_filter();
      break;

    case HCI_SET_EVENTS_STATE:
      if (hci_cmd_complete || noresp_limit) {
        hci_write_authentication_enable(0 < strlen(&btdPin[0]));
      }
      break;

    case HCI_WRITE_AUTHENTICATION_ENABLE:
      if (hci_cmd_complete) hci_write_pin_type(true); // True - Variable PIN
      break;

    case HCI_WRITE_PIN_TYPE:
      if (hci_cmd_complete) {
        memset( disc_bdaddr, 0, sizeof(disc_bdaddr) );
        hci_read_bdaddr();
      }
      break;

    case HCI_BDADDR_STATE:
      if (hci_read_bdaddr_complete || noresp_limit) {
#ifdef DEBUG_HCI_EVENTS
        printf("%lu - Local Bluetooth Address: ", ticks);
        for (int8_t i = 5; i > 0; i--) {
          printf("%02.2x:",my_bdaddr[i]);
        }
        printf("%02.2x\r\n",my_bdaddr[0]);
#endif
        hci_read_local_version_information();
      }
      break;

    case HCI_LOCAL_VERSION_STATE:
      if (hci_read_version_complete) {
        if (0 < strlen(&(btdName[0]))) {
          hci_set_local_name(&(btdName[0]));
        }
        else if (bEirCapable) {
          hci_write_extended_inquiry_response();
        }
        else {
          hci_write_page_timeout();
        }
      }
      break;

    case HCI_SET_NAME_STATE:
      if (hci_cmd_complete) {
        if ( bEirCapable ) {
          hci_write_extended_inquiry_response();
        }
        else {
          hci_write_page_timeout();
        }
      }
      break;

    case HCI_EXTENDED_INQUIRY_RESPONSE_SENT_STATE:
      if (hci_cmd_complete) {
        hci_write_page_timeout();
      }
      break;

    case HCI_WRITE_PAGE_TIMEOUT:
      if (hci_cmd_complete) hci_write_connection_accept_timeout();
      break;

   case HCI_WRITE_CONNECTION_ACCEPT_TIMEOUT:
      if (hci_cmd_complete) hci_write_scan_enable();
      break;

    case HCI_SCANNING_STATE:
      printf("%lu - Wait For Incoming Connection Request. Band: %u\r\n", ticks, sizeof(uhd_l2cap_queue_t));
      connectionWait = true;
      // hci_read_stored_link_key(NULL); // Trigger reading of all link keys
      hci_state = HCI_CONNECT_IN_STATE;
      break;

    case HCI_CONNECT_IN_STATE:
      if (hci_incoming_connect_request) {
        connectionWait = false;
        printf("%lu - Incoming Connection Request\r\n", ticks);
        hci_accept_connection();
      }
      break;

    case HCI_ACCEPT_CONNECTION_STATE:
      // At this point, no authentication has occurred yet.
      // Wait for completion, no-op here. Next state is HCI_HANDLE_RECEIVED
      break;

    case HCI_SIMPLE_PAIRING_COMPLETE:
      // hci_connect();
      break;

    case HCI_LINK_KEY_NOTIFICATION_EVENT:
      if (hci_link_key_notification_event) {
        printf("%lu - ! Link key notification event following [ 0x%08.8X ]\r\n", ticks, hci_last_invoked);
        hci_link_key_notification();
      }
      break;

    case HCI_LINK_KEY_NOTIFICATION_HANDLED:
      // Holding state
      break;

    case HCI_HANDLE_RECEIVED:
      if (hci_connect_complete) {
        // hci_change_connection_packet_type();
        hci_remote_name();
        // hci_set_flow_specification();
      }
      break;

    case HCI_CHANGE_CONNECTION_PACKET_TYPE:
      if (hci_change_connection_packet_type_event) {
        hci_remote_name();
      }
      break;

    case HCI_FLOW_CONTROL_ENABLE:
      if (hci_flow_specification_event) {
        printf("%lu - HCI Flow Specification Event Received.\r\n", ticks);
        // Reassert flag
        // hci_event_flag |= HCI_FLAG_CONN_COMPLETE;
        // hci_state = HCI_CONNECTED_STATE;
        hci_remote_name();
      }
      break;

    case HCI_REMOTE_NAME_STATE:
      if (hci_remote_name_complete) {
        printf("%lu - Remote Name: ", ticks);
        for (uint8_t i = 0; i < 30; i++) {
          if (remote_name[i] == NULL)
            break;
          printf("%c",remote_name[i]);
        }
        printf("\r\n");
        // hci_set_flow_specification();
#ifdef CHARACTERIZE_REMOTE_HOST
        hci_state = HCI_CHARACTERIZE_REMOTE;
#else
        hci_event_flag |= HCI_FLAG_CONN_COMPLETE;
        hci_state = HCI_CONNECTED_STATE;
#endif
      }
      break;

    ////////////////////////////////////////////////////////////////
    /////////////   Asynchronous events  ///////////////////////////
    ////////////////////////////////////////////////////////////////

    case HCI_USER_CONFIRMATION_REQUEST:
      // Default reply is YES
      printf("%lu - Sending affirmative response to confirmation request\r\n", ticks);
      hci_user_confirmation_request_reply();
      break;

    case HCI_USER_CONFIRMATION_REQUEST_REPLY:
      if (hci_cmd_complete) {
        // hci_state = HCI_ASYNC_HOLD; 
        hci_state = HCI_CONNECT_STATE; 
      }
      break;

    case HCI_CONNECT_STATE:
      break;

    case HCI_CONNECTED_STATE:
      if (hci_connect_complete) {
        printf("%lu - Connected to Device: ", ticks);
        for (int8_t i = 0; i < 6; i++) {
          printf("%02.2x:",disc_bdaddr[5-i]);
        }
        printf("\r\n");
        // Clear these flags for a new connection
        l2capConnectionClaimed  = false;
        sdpConnectionClaimed    = false;
        rfcommConnectionClaimed = false;

        hci_event_flag = 0;
        hci_counter    = 0;
      }
      break;

    case HCI_CHARACTERIZE_REMOTE:
      hci_read_remote_supported_features();
      break;

    case HCI_READ_REMOTE_SUPPORTED_FEATURES:
      if (hci_read_remote_supported_features_complete_event) {
        // hci_read_remote_extended_features();
        hci_read_remote_version_information();
      }
      break;

    case HCI_READ_REMOTE_EXTENDED_FEATURES:
      if (hci_read_remote_extended_features_complete_event) {
        hci_read_remote_version_information();
      }
      break;

    case HCI_READ_REMOTE_VERSION_INFORMATION:
      // hci_state = HCI_ASYNC_HOLD;
      // hci_state = HCI_ACL_U_SETUP;
      if (hci_read_remote_version_information_complete_event) {
        hci_set_connection_encryption();
      }
      break;

    case HCI_ENCRYPTION_CHANGE:
      if (hci_encryption_change_event) {
        hci_state = HCI_CONNECTED_STATE;
      }
      break;

    case HCI_DONE_STATE:
    case HCI_ASYNC_HOLD:
      // Hold states.
      break;

    case HCI_ACL_U_SETUP:
      hci_state = HCI_ACL_U_READY;
      break;

    case HCI_ACL_U_READY:
      // Hold here
      // hci_state = HCI_CONNECTED_STATE;
      break;

    case HCI_DISCONNECT_STATE:
      if (hci_disconnect_complete) {
        printf("%lu - HCI Disconnected from Device\r\n", ticks);
        hci_event_flag = 0; // Clear all flags
        // Reset all buffers
        memset(hcibuf, 0, sizeof(hcibuf));
        memset(disc_bdaddr, 0, sizeof(disc_bdaddr));
        hci_state = HCI_SCANNING_STATE;
      }
      break;

    case HCI_IO_CAPABILITY_REQUESTED_STATE:
      if (hci_io_capability_request_event) {
        hci_io_capability_response();
      }
      break;

    case HCI_IO_CAPABILITY_RESPONSE_SENT_STATE:
      // Holding state
      break;
      
    case HCI_OOB_DATA_REQUEST_STATE:
      if ( hci_oob_data_request_event ) {
        hci_oob_data_request_reply();
      }
      break;

    case HCI_OOB_DATA_REQUEST_REPLY_SENT_STATE:
      break;

    default:
      break;
  }/*}}}*/
  pUsb->setHciState(hci_state);
}/*}}}*/

/************************************************************/
/*                    HCI Commands                          */
/************************************************************/

void BTD::hci_dump_event() {/*{{{*/
#ifdef DEBUG_HCI_EVENTS
  uint32_t event_payload_len = hcibuf[1];
  uint32_t dump_limit = min(sizeof(hcibuf), event_payload_len + 2);
  uint32_t eff_limit = min(sizeof(hcibuf), dump_limit + (8 - (dump_limit % 8))); 
  printf("\r\n-- EV 0x%02.2X", hcibuf[0]);
  for ( uint32_t b = 0 ; b < eff_limit ; b++ ) {
    if ( !( b % 8 ) ) printf("\r\n");
    printf("%02.2X%c", hcibuf[b], b < dump_limit ? ' ' : '.');
  }
  printf("\r\n\n");
#endif
}/*}}}*/

void BTD::hci_dump_command() {/*{{{*/
  uint32_t event_payload_len = hcicmd[2];
  uint32_t dump_limit = min(sizeof(hcicmd), event_payload_len + 3);
  uint32_t eff_limit = min(sizeof(hcicmd), dump_limit + (8 - (dump_limit % 8))); 
  printf("\r\n-- OCF(0x%02.2X) OGF(0x%02.2X) Len(%u)", hcicmd[0], hcicmd[1], hcicmd[2]);
  for ( uint32_t b = 0 ; b < eff_limit ; b++ ) {
    if ( !( b % 8 ) ) printf("\r\n");
    printf("%02.2X%c", hcicmd[b], b < dump_limit ? ' ' : '.');
  }
  printf("\r\n\n");
}/*}}}*/

void BTD::hci_dump_completed_packets_event(uint8_t *event_data) {/*{{{*/
  printf("Completed packets: %u\r\n", event_data[2]);
  for ( uint8_t i = 0 ; i < event_data[2] ; i++ ) {
    uint8_t offset = 3 + (i * 4); 
    printf("- CH 0x%02.2X%02.2X: %u\r\n", event_data[offset+1], event_data[offset], event_data[offset+2] | event_data[offset+3] << 8 );
  }
}/*}}}*/

void HCI_Command_(usb_add_t add, uhd_trans_status_t status, uint16_t payload_trans)
{/*{{{*/
  // Update state
  btd_instance->HCI_event_task(status);
}/*}}}*/

bool BTD::HCI_Command(uint8_t* data, uint16_t nbytes, uint16_t read_length) {/*{{{*/
  // Use the control pipe to send HCI-specific commands
  // (Core Spec 4.0 Vol. 4 Part B Sec. 2.2)
  uint32_t timeout = GetTickCount() + 20; 

  usb_setup_req_t req;

  req.bmRequestType = bmREQ_HCI_OUT;
  req.bRequest      = 0;
  req.wValue        = 0; 
  req.wIndex        = 0;
  req.wLength       = nbytes;  

  hci_event_flag &= ~HCI_FLAG_CMD_COMPLETE;

  if( !uhd_setup_request(bAddress, &req, (uint8_t *)data, nbytes, NULL, &HCI_Command_ )) {/*{{{*/
    // Caller need only know that we failed to issue the HCI command over 
    // the pipe.
    TRACE_USBHOST_HCI(printf("--> HCI %u-byte command to device 0x%02.2x\r\n",nbytes,bAddress);)
    hci_dump_command();
    return false; 
  }/*}}}*/
  return true;
}/*}}}*/

void BTD::hci_read_local_supported_commands() {/*{{{*/

  hci_event_flag &= ~HCI_FLAG_CMD_COMPLETE;
  hci_state = HCI_READ_LOCAL_SUPPORTED_COMMANDS;

  hcicmd[0] = 0x02; // HCI OCF = 0x0002, Read Local Supported Commands 
  hcicmd[1] = 0x04 << 2; // HCI OGF = 4 (Control and Baseband, Core Rev 4.0 Vol 2.E.7.4)
  hcicmd[2] = 0x00; // No parameters 

  hci_last_invoked = &BTD::hci_read_local_supported_commands;
  TRACE_USBHOST_HCI(printf("%lu - Executed read_local_supported_commands [ 0x%08.8X ]\r\n", ticks, hci_last_invoked);)
  HCI_Command(hcicmd, 3, 65);
}/*}}}*/

void BTD::hci_read_local_supported_features() {/*{{{*/

  hci_event_flag &= ~HCI_FLAG_CMD_COMPLETE;
  hci_state = HCI_READ_LOCAL_SUPPORTED_FEATURES;

  hcicmd[0] = 0x03; // HCI OCF = 0x0002, Read Local Supported Commands 
  hcicmd[1] = 0x04 << 2; // HCI OGF = 3 (Control and Baseband, Core Rev 4.0 Vol 2.E.7.4)
  hcicmd[2] = 0x00; // No parameters 

  hci_last_invoked = &BTD::hci_read_local_supported_features;
  TRACE_USBHOST_HCI(printf("%lu - Executed read_local_supported_features [ 0x%08.8X ]\r\n", ticks, hci_last_invoked);)
  HCI_Command(hcicmd, 3, 65);
}/*}}}*/

void BTD::hci_set_default_event_mask() {/*{{{*/
  hci_state  = HCI_SET_EVTMASK_STATE;
  hcicmd[0]  = 0x01; // HCI OCF = 0x0001, Set event mask 
  hcicmd[1]  = 0x03 << 2; // HCI OGF = 3 (Control and Baseband, Core Rev 4.0 Vol 2.E.7.3)
  hcicmd[2]  = 0x08;
  // All flags: 0x 3D BF F8 07 FF FB 9F FF
  hcicmd[3]  = 0xFF;
  hcicmd[4]  = 0x9F;
  hcicmd[5]  = 0xFB;
  hcicmd[6]  = 0xFF;

  hcicmd[7]  = 0x07;
  hcicmd[8]  = 0xF8;
  hcicmd[9]  = 0xBF;
  hcicmd[10] = 0x3D;

  TRACE_USBHOST_HCI(printf("%lu - Setting event mask 0x%02.2X 0x%02.2X 0x%02.2X 0x%02.2X 0x%02.2X 0x%02.2X 0x%02.2X 0x%02.2X\r\n", ticks, hcicmd[3], hcicmd[4], hcicmd[5], hcicmd[6], hcicmd[7], hcicmd[8], hcicmd[9], hcicmd[10]);)

  HCI_Command(hcicmd, 11);
  hci_last_invoked = &BTD::hci_set_default_event_mask;
  TRACE_USBHOST_HCI(printf("%lu - Executed default_event_mask [ 0x%08.8X ]\r\n", ticks, hci_last_invoked);)
}/*}}}*/

void BTD::hci_clear_event_filter() {/*{{{*/
  TRACE_USBHOST_HCI(printf("%lu - Clearing event filters. [ 0x%08.8X ]\r\n", ticks, hci_last_invoked);)
  hci_set_event_filter( EV_FILTER_TYPE_CLEAR_ALL );
  hci_state = HCI_CLEAR_EVENTS_STATE;
  hci_last_invoked = &BTD::hci_clear_event_filter;
  TRACE_USBHOST_HCI(printf("%lu - Executed clear_event_filter [ 0x%08.8X ]\r\n", ticks, hci_last_invoked);)
}/*}}}*/

void BTD::hci_set_default_event_filter() {/*{{{*/
  TRACE_USBHOST_HCI(printf("%lu - Setting default event filters. [ 0x%08.8X ]\r\n", ticks, hci_last_invoked);)
  uint8_t accept_all[1];
  accept_all[0] = EV_FILTER_COND_CONN_ALL_WITH_ROLESWITCH;
  hci_set_event_filter(
    EV_FILTER_TYPE_CONNECTION_SETUP,
    EV_FILTER_COND_CONN_ACCEPT_ALL,
    1,
    &(accept_all[0])
  );
  hci_state = HCI_SET_EVENTS_STATE;
  hci_last_invoked = &BTD::hci_set_default_event_filter;
  TRACE_USBHOST_HCI(printf("%lu - Executed set_event_default_filter [ 0x%08.8X ]\r\n", ticks, hci_last_invoked);)
}/*}}}*/

void BTD::hci_set_event_filter(uint8_t filter_type, uint8_t filter_cond, uint8_t param_len, uint8_t * filter_params ) {/*{{{*/
  // Set event filter, see Bluetooth Specification V 4.0, Vol. 2, Part E, Sec. 7.3.3.
  hci_state = HCI_SET_EVENTS_STATE;
  hcicmd[0] = 0x05; // HCI OCF = 0x0005
  hcicmd[1] = 0x03 << 2; // HCI OGF = 3 (Control and Baseband, Core Rev 4.0 Vol 2.E.7.3)
  hcicmd[2] = 1;
  hcicmd[3] = filter_type;
  uint16_t len = 0;
  switch ( filter_type ) {
    case EV_FILTER_TYPE_CLEAR_ALL: 
      len = 4;
      break;
    case EV_FILTER_TYPE_INQUIRY_RESULT:
      hcicmd[4] = filter_cond;
      switch ( filter_cond ) {/*{{{*/
        case EV_FILTER_COND_INQ_ACCEPT_ALL:
          len = 5;
          break;
        case EV_FILTER_COND_INQ_CLASS_RESPONDER:
        case EV_FILTER_COND_INQ_ADDR_RESPONDER:
        default: return;
      }/*}}}*/
      break;
    case EV_FILTER_TYPE_CONNECTION_SETUP:
      hcicmd[4] = filter_cond;
      switch ( filter_cond ) {/*{{{*/
        case EV_FILTER_COND_CONN_ACCEPT_ALL:
          if ( param_len == 1 ) {
            hcicmd[5] = filter_params[0]; // AUTO_ACCEPT flag
            len = 6;
          } else return;
          break;
        case EV_FILTER_COND_CONN_CLASS_MATCHED:
          if ( param_len == 7 ) {
            len = 5;
            while ( param_len ) {
              len++;
              hcicmd[len + param_len - 1] = filter_params[param_len-1];
              param_len--;
            }
          } else return;
        case EV_FILTER_COND_CONN_ADDR_MATCHED:
        default: return;
      }/*}}}*/
      break;
    default: return;
  }
  hci_last_invoked = NULL; 
  HCI_Command(hcicmd, len);
}/*}}}*/

void BTD::hci_reset() {/*{{{*/
  hci_event_flag = 0; // Clear all the flags
  hci_state = HCI_RESET_STATE;
  hcicmd[0] = 0x03; // HCI_Reset Vol 4.2.E.7.3.2 
  hcicmd[1] = 0x03 << 2; // HCI OGF = 3 (Control and Baseband, Core Rev 4.0 Vol 2.E.7.3)
  hcicmd[2] = 0x00; // parameter length = 0

  HCI_Command(hcicmd, 3, 1);
  hci_last_invoked = &BTD::hci_reset;
  TRACE_USBHOST_HCI(printf("%lu - Executed reset. [ 0x%08.8X ]\r\n", ticks, hci_last_invoked);)
}/*}}}*/

void BTD::hci_write_scan_enable() {/*{{{*/
  hci_event_flag &= ~(HCI_FLAG_INCOMING_REQUEST | HCI_FLAG_CMD_COMPLETE);
  hci_state = HCI_SCANNING_STATE;
  hcicmd[0] = 0x1A; // HCI OCF = 1A
  hcicmd[1] = 0x03 << 2; // HCI OGF = 3
  hcicmd[2] = 0x01; // parameter length = 1
  if (0 < strlen(btdName))
    hcicmd[3] = 0x03; // Inquiry Scan enabled. Page Scan enabled.
  else
    hcicmd[3] = 0x02; // Inquiry Scan disabled. Page Scan enabled.

  HCI_Command(hcicmd, 4, 1);
  hci_last_invoked = &BTD::hci_write_scan_enable;
  TRACE_USBHOST_HCI(printf("%lu - Executed write_scan_enable. [ 0x%08.8X ]\r\n", ticks, hci_last_invoked);)
}/*}}}*/

void BTD::hci_write_connection_accept_timeout() {/*{{{*/

  hci_event_flag &= ~HCI_FLAG_CMD_COMPLETE;
  hci_state = HCI_WRITE_CONNECTION_ACCEPT_TIMEOUT;

  hcicmd[0] = 0x16; // HCI OCF = 1A
  hcicmd[1] = 0x03 << 2; // HCI OGF = 3
  hcicmd[2] = 2;
  hcicmd[3] = 0xE0; // Two octets, count of baseband slots @ 0.625 msec
  hcicmd[4] = 0x2E; // 7.5 sec: 0x2EE0 (12,000) 

  HCI_Command(hcicmd, 5);
  hci_last_invoked = &BTD::hci_write_connection_accept_timeout;
  TRACE_USBHOST_HCI(printf("%lu - Executed write_connection_accept_timeout. [ 0x%08.8X ]\r\n", ticks, hci_last_invoked);)

}/*}}}*/

void BTD::hci_write_page_timeout() {/*{{{*/

  hci_event_flag &= ~HCI_FLAG_CMD_COMPLETE;
  hci_state = HCI_WRITE_PAGE_TIMEOUT;

  hcicmd[0] = 0x18; // HCI OCF = 1A
  hcicmd[1] = 0x03 << 2; // HCI OGF = 3
  hcicmd[2] = 2;
  hcicmd[3] = 0xE0; // Two octets, count of baseband slots @ 0.625 msec
  hcicmd[4] = 0x2E; // 7.5 sec: 0x2EE0 (12,000) 

  HCI_Command(hcicmd, 5);
  hci_last_invoked = &BTD::hci_write_page_timeout;
  TRACE_USBHOST_HCI(printf("%lu - Executed write_page_timeout. [ 0x%08.8X ]\r\n", ticks, hci_last_invoked);)

}/*}}}*/

void BTD::hci_write_scan_disable() {/*{{{*/

  hci_event_flag &= ~HCI_FLAG_CMD_COMPLETE;
  hcicmd[0] = 0x1A; // HCI OCF = 1A
  hcicmd[1] = 0x03 << 2; // HCI OGF = 3
  hcicmd[2] = 0x01; // parameter length = 1
  hcicmd[3] = 0x00; // Inquiry Scan disabled. Page Scan disabled.

  HCI_Command(hcicmd, 4);
  hci_last_invoked = &BTD::hci_write_scan_disable;
  TRACE_USBHOST_HCI(printf("%lu - Executed write_scan_disable. [ 0x%08.8X ]\r\n", ticks, hci_last_invoked);)
}/*}}}*/

void BTD::hci_write_simple_pairing_mode(bool v) {/*{{{*/
  hci_state = HCI_WRITE_SIMPLE_PAIRING_MODE_STATE;
  hci_event_flag &= ~HCI_FLAG_CMD_COMPLETE;
  hcicmd[0] = 0x56; // HCI OCF = 0x56 
  hcicmd[1] = 0x03 << 2; // HCI OGF = 3
  hcicmd[2] = 0x01; // parameter length = 1
  hcicmd[3] = v ? 0x01 : 0x00; 

  HCI_Command(hcicmd, 4);
  hci_last_invoked = NULL;
  TRACE_USBHOST_HCI(printf("%lu - Executed write_simple_pairing_mode(%s). [ 0x%08.8X ]\r\n", ticks, bSspCapable ? "true" : "false", hci_last_invoked);)
}/*}}}*/

void BTD::hci_read_bdaddr() {/*{{{*/
  hci_event_flag &= ~HCI_FLAG_CMD_COMPLETE;
  hci_state = HCI_BDADDR_STATE;
  hcicmd[0] = 0x09; // HCI OCF = 9
  hcicmd[1] = 0x04 << 2; // HCI OGF = 4
  hcicmd[2] = 0x00;

  HCI_Command(hcicmd, 3, 7);
  hci_last_invoked = &BTD::hci_read_bdaddr;
  TRACE_USBHOST_HCI(printf("%lu - Executed read_bdaddr. [ 0x%08.8X ]\r\n", ticks, hci_last_invoked);)
}/*}}}*/

void BTD::hci_read_local_version_information() {/*{{{*/
  hci_state = HCI_LOCAL_VERSION_STATE;
  hcicmd[0] = 0x01; // HCI OCF = 1
  hcicmd[1] = 0x04 << 2; // HCI OGF = 4
  hcicmd[2] = 0x00;

  HCI_Command(hcicmd, 3, 9);
  hci_last_invoked = &BTD::hci_read_local_version_information;
  TRACE_USBHOST_HCI(printf("%lu - Executed read_local_version_information. [ 0x%08.8X ]\r\n", ticks, hci_last_invoked);)
}/*}}}*/

void BTD::hci_remote_name() {/*{{{*/
  hci_event_flag &= ~HCI_FLAG_REMOTE_NAME_COMPLETE;
  hci_state = HCI_REMOTE_NAME_STATE;
  hcicmd[0] = 0x19; // HCI OCF = 19
  hcicmd[1] = 0x01 << 2; // HCI OGF = 1
  hcicmd[2] = 0x0A; // parameter length = 10
  hcicmd[3] = disc_bdaddr[0]; // 6 octet bdaddr
  hcicmd[4] = disc_bdaddr[1];
  hcicmd[5] = disc_bdaddr[2];
  hcicmd[6] = disc_bdaddr[3];
  hcicmd[7] = disc_bdaddr[4];
  hcicmd[8] = disc_bdaddr[5];
  hcicmd[9] = 0x01; //Page Scan Repetition Mode
  hcicmd[10] = 0x00; //Reserved
  hcicmd[11] = 0x00; //Clock offset - low byte
  hcicmd[12] = 0x00; //Clock offset - high byte

  HCI_Command(hcicmd, 13);
  hci_last_invoked = &BTD::hci_remote_name;
  TRACE_USBHOST_HCI(printf("%lu - Executed remote_name. [ 0x%08.8X ]\r\n", ticks, hci_last_invoked);)
}/*}}}*/

void BTD::hci_write_extended_inquiry_response() {/*{{{*/

  hci_event_flag &= ~HCI_FLAG_CMD_COMPLETE;
  hci_state = HCI_EXTENDED_INQUIRY_RESPONSE_SENT_STATE;

  uint32_t cmdlen = 4;

  hcicmd[0] = 0x52;
  hcicmd[1] = 0x03 << 2;
  hcicmd[2] = 0xF1; // parameter length = 241, including FEC flag 
  hcicmd[3] = 0x00; // FEC Required, boolean 0x00 / 0x01

  memset( &(hcicmd[4]), 0, 240 );

	for ( uint32_t i = 0 ; i < sizeof(arduino_due_btserial) ; i++ ) {
		hcicmd[i+4] = pgm_read_byte(&arduino_due_btserial[i]);
    cmdlen++;
	}

  cmdlen = 244;

  HCI_Command(hcicmd, cmdlen);

  hci_last_invoked = &BTD::hci_write_extended_inquiry_response;
  TRACE_USBHOST_HCI(printf("%lu - Executed write_extended_inquiry_response, cmdlen %lu [ 0x%08.8X ]\r\n", ticks, cmdlen, hci_last_invoked);)
}/*}}}*/

void BTD::hci_set_local_name(const char* name) {/*{{{*/
  hci_state = HCI_SET_NAME_STATE;
  hcicmd[0] = 0x13; // HCI OCF = 13
  hcicmd[1] = 0x03 << 2; // HCI OGF = 3
  hcicmd[2] = strlen(name) + 1; // parameter length = the length of the string + end byte
  uint8_t i;
  for (i = 0; i < strlen(name); i++)
    hcicmd[i + 3] = name[i];
  hcicmd[i + 3] = 0x00; // End of string

  HCI_Command(hcicmd, 4 + strlen(name));
  hci_last_invoked = NULL; 
  TRACE_USBHOST_HCI(printf("%lu - Executed set_local_name: '%s'\r\n", ticks, name);)
}/*}}}*/

void BTD::hci_inquiry() {/*{{{*/
  hci_state = HCI_INQUIRY_STATE; 
  hci_event_flag &= ~HCI_FLAG_NONEMPTY_INQUIRY_RESULT;
  hcicmd[0] = 0x01;
  hcicmd[1] = 0x01 << 2; // HCI OGF = 1
  hcicmd[2] = 0x05; // Parameter Total Length = 5
  hcicmd[3] = 0x33; // LAP: General/Unlimited Inquiry Access Code (GIAC = 0x9E8B33) - see https://www.bluetooth.org/Technical/AssignedNumbers/baseband.htm
  hcicmd[4] = 0x8B;
  hcicmd[5] = 0x9E;
  hcicmd[6] = 0x30; // Inquiry time N * 1.28s; 48 => 61.44 sec (maximum)
  hcicmd[7] = 0x0A; // 0x0A: 10 number of responses; 0x00: Unlimited responses

  HCI_Command(hcicmd, 8);
  hci_last_invoked = &BTD::hci_inquiry;
  TRACE_USBHOST_HCI(printf("%lu - Executed inquiry. [ 0x%08.8X ]\r\n", ticks, hci_last_invoked);)
}/*}}}*/

void BTD::hci_inquiry_cancel() {/*{{{*/
  hcicmd[0] = 0x02;
  hcicmd[1] = 0x01 << 2; // HCI OGF = 1
  hcicmd[2] = 0x00; // Parameter Total Length = 0

  HCI_Command(hcicmd, 3);
  hci_last_invoked = &BTD::hci_inquiry_cancel;
  TRACE_USBHOST_HCI(printf("%lu - Executed inquiry_cancel. [ 0x%08.8X ]\r\n", ticks, hci_last_invoked);)
}/*}}}*/

void BTD::hci_accept_connection() {/*{{{*/

  hci_event_flag &= ~HCI_FLAG_CONN_COMPLETE;
  hci_state = HCI_ACCEPT_CONNECTION_STATE; // FIXME: Enter holding state

  hcicmd[0] = 0x09; // HCI OCF = 9
  hcicmd[1] = 0x01 << 2; // HCI OGF = 1
  hcicmd[2] = 0x07; // parameter length 7
  memmove( &(hcicmd[3]), disc_bdaddr, 6 );
  hcicmd[9] = 0x01; // 0x00, Switch role to master; 0x01 remain slave

  HCI_Command(hcicmd, 10);
  hci_last_invoked = &BTD::hci_accept_connection;
  TRACE_USBHOST_HCI(printf("%lu - Executed accept_connection. [ 0x%08.8X ]\r\n", ticks, hci_last_invoked);)
}/*}}}*/

void BTD::hci_set_controller_to_host_flow_control() {/*{{{*/
  hci_event_flag &= ~HCI_FLAG_CMD_COMPLETE;
  hci_state = HCI_C_TO_H_FLOW_CONTROL_ENABLE; // FIXME: Enter holding state

  hcicmd[0] = 0x31; // OCF 0x31
  hcicmd[1] = 0x03 << 2; // HCI OGF
  hcicmd[2] = 0x01; // parameter length 1
  hcicmd[3] = 0x01; // Bit 0: ACL; bit 1: Synchronous HCI packets

  HCI_Command(hcicmd, 4);

  hci_last_invoked = &BTD::hci_set_controller_to_host_flow_control;
  TRACE_USBHOST_HCI(printf("%lu - Executed set_controller_to_host_flow_control. [ 0x%08.8X ]\r\n", ticks, hci_last_invoked);)

}/*}}}*/

void BTD::hci_read_host_buffer_size() {/*{{{*/

  hci_event_flag &= ~HCI_FLAG_CMD_COMPLETE;
  hci_state = HCI_READ_HOST_BUFFER_SIZE; // FIXME: Enter holding state

  hcicmd[0] = 0x05; // OCF 0x05 7.4.5
  hcicmd[1] = 0x04 << 2; // HCI OGF
  hcicmd[2] = 0x00; // parameter length

  HCI_Command(hcicmd, 3);

  hci_last_invoked = &BTD::hci_read_host_buffer_size;
  TRACE_USBHOST_HCI(printf("%lu - Executed read_host_buffer_size. [ 0x%08.8X ]\r\n", ticks, hci_last_invoked);)

}/*}}}*/

void BTD::hci_set_host_buffer_size() {/*{{{*/
  /*
   * The Host_Buffer_Size command is used by the Host to notify the Controller
   * about the maximum size of the data portion of HCI ACL and synchronous
   * Data Packets sent from the Controller to the Host. The Controller shall
   * segment the data to be transmitted from the Controller to the Host
   * according to these sizes, so that the HCI Data Packets will contain 
   * data with up to these sizes. The Host_Buffer_Size command also notifies 
   * the Controller about the total number of HCI ACL and synchronous 
   * Data Packets that can be stored in the data buffers of the Host. 
   * If flow control from the Controller to the Host is turned off, 
   * and the Host_Buffer_Size command has not been issued by the Host, 
   * this means that the Controller will send HCI Data Packets to the Host 
   * with any lengths the Controller wants to use, and it is assumed that 
   * the data buffer sizes of the Host are unlimited.
   *
   * If flow control from the Controller to the Host is turned on, the
   * Host_Buffer_Size command shall after a power-on or a reset always be
   * sent by the Host before the first Host_Number_Of_Completed_Packets 
   * command is sent.
   */
  hci_event_flag &= ~HCI_FLAG_CMD_COMPLETE;
  hci_state = HCI_SET_HOST_BUFFER_SIZE; // FIXME: Enter holding state

  hcicmd[0] = 0x33; // OCF 0x33 7.3.39
  hcicmd[1] = 0x03 << 2; // HCI OGF
  hcicmd[2] = 0x07; // parameter length
  hcicmd[3] = 0x36; //0x36; // HCI ACL Data Packet length (default 310 = 0x0136)
  hcicmd[4] = 0x01; //0x01;
  hcicmd[5] = 0x40; // HCI Synchronous Data Packet length (default 64 = 0x40)
  hcicmd[6] = 0x0A; // Maximum count of HCI ACL data packets cacheable (default 10 = 0x000A)
  hcicmd[7] = 0x00; 
  hcicmd[8] = 0x08; // Maximum count of HCI synchronous data packets cacheable (default 8 = 0x0008)
  hcicmd[9] = 0x00;

  HCI_Command(hcicmd, 10);

  hci_last_invoked = &BTD::hci_set_host_buffer_size;
  TRACE_USBHOST_HCI(printf("%lu - Executed set_host_buffer_size. [ 0x%08.8X ]\r\n", ticks, hci_last_invoked);)

}/*}}}*/

void BTD::hci_host_number_of_packets_completed(uint16_t packets) {/*{{{*/

  // See Core Spec 4.0 Vol 2, Part E.7.3.40 (Host Number Of Completed Packets Command)
  // Also Vol 2, Part E.7.7.19 (Number Of Completed Packets Event)
  // - EV_NUM_COMPLETE_PKT (0x13)
  // triggered by availability of HCI DATA pkts from remote host
  hcicmd[0] = 0x35; // OCF
  hcicmd[1] = 0x03 << 2; // OGF
  hcicmd[2] = 5; // parameter length
  hcicmd[3] = 1; // Number of handles (just the one we have)
  hcicmd[4] = (uint8_t)(hci_handle & 0xFF);
  hcicmd[5] = (uint8_t)((hci_handle >> 8) & 0x0F);
  hcicmd[6] = (uint8_t)(packets & 0xFF);
  hcicmd[7] = (uint8_t)((packets >> 8) & 0xFF);

  HCI_Command(hcicmd, 8);
}/*}}}*/

void BTD::hci_set_flow_specification() {/*{{{*/
  // This command may result in a Command Status (4.2.E.7.7.15) event 0x0F being sent to the Due.
  // This command may result in a Flow Specification Complete (4.2.E.7.7.32) event 0x21  being sent to the Due.
  hci_event_flag &= ~(HCI_FLAG_CMD_COMPLETE | HCI_FLAG_FLOW_SPECIFICATION);
  hci_state = HCI_FLOW_CONTROL_ENABLE; // FIXME: Enter holding state

  // See Core Spec 4.0 Vol 2, Part E.7.2.13
  hcicmd[0] = 0x10; // OCF 0x10
  hcicmd[1] = 0x02 << 2; // HCI OGF
  hcicmd[2] = 21; // parameter length
  hcicmd[3] = (uint8_t)(hci_handle & 0xFF); //connection handle - low byte
  hcicmd[4] = (uint8_t)((hci_handle >> 8) & 0x0F); //connection handle - high byte
  hcicmd[5] = 0x00; // Reserved
  hcicmd[6] = 0x01; // Flow_Direction: 0x01 - Incoming, 0x00 - Outgoing
  hcicmd[7] = 0x01; // Service_Type: 0x01 - Best Effort, 0x02 - Guaranteed
  hcicmd[8] = 0x00; // Token Rate: 4 octets, rate in octets / second. Don't care: 0x00000000
  hcicmd[9] = 0x00;
  hcicmd[10] = 0x00;
  hcicmd[11] = 0x00;
  hcicmd[12] = 0xFF; // Token bucket size, octets. Don't care: 0x00000000
  hcicmd[13] = 0xFF;
  hcicmd[14] = 0xFF;
  hcicmd[15] = 0xFF;
  hcicmd[16] = 0x00; // Peak bandwidth. Don't care: 0x00000000 
  hcicmd[17] = 0x00;
  hcicmd[18] = 0x00;
  hcicmd[19] = 0x00;
  hcicmd[20] = 0xFF; // Access latency. Don't care: 0xFFFFFFFF 
  hcicmd[21] = 0xFF;
  hcicmd[22] = 0xFF;
  hcicmd[23] = 0xFF;

  // Force successful completion when the HCI command fails.
  if ( !HCI_Command(hcicmd, 24) ) hci_event_flag |= HCI_FLAG_FLOW_SPECIFICATION;

  hci_last_invoked = &BTD::hci_set_flow_specification;
  TRACE_USBHOST_HCI(printf("%lu - Executed set_flow_specification. [ 0x%08.8X ]. Expect Command Status event.\r\n", ticks, hci_last_invoked);)

}/*}}}*/

void BTD::hci_user_confirmation_request_reply() {/*{{{*/
  hci_event_flag &= ~HCI_FLAG_CMD_COMPLETE;
  hci_state = HCI_USER_CONFIRMATION_REQUEST_REPLY; // FIXME: Enter holding state

  hcicmd[0] = 0x2C; // OCF
  hcicmd[1] = 0x01 << 2; // HCI OGF = 1
  hcicmd[2] = 0x06;
  memmove( &(hcicmd[3]), disc_bdaddr, 6 );

  HCI_Command(hcicmd, 9);
  hci_last_invoked = &BTD::hci_user_confirmation_request_reply;
  TRACE_USBHOST_HCI(printf("%lu - Executed user_confirmation_request_reply. [ 0x%08.8X ]\r\n", ticks, hci_last_invoked);)
}/*}}}*/

void BTD::hci_connect() {/*{{{*/

  hci_state = HCI_CONNECT_STATE; 
  hci_event_flag &= ~(HCI_FLAG_CONN_COMPLETE | HCI_FLAG_CONNECT_EVENT);

  hcicmd[0] = 0x05;
  hcicmd[1] = 0x01 << 2; // HCI OGF = 1
  hcicmd[2] = 0x0D; // parameter Total Length = 13
  memmove( &(hcicmd[3]), disc_bdaddr, 6 );
  hcicmd[9] = 0x18; // DM1 or DH1 may be used
  hcicmd[10] = 0xCC; // DM3, DH3, DM5, DH5 may be used
  hcicmd[11] = 0x01; // Page repetition mode R1
  hcicmd[12] = 0x00; // Reserved
  hcicmd[13] = 0x00; // Clock offset
  hcicmd[14] = 0x00; // Invalid clock offset
  hcicmd[15] = 0x01; // Role switch: 1 = permit, 0 = remain master

  HCI_Command(hcicmd, 16);
  hci_last_invoked = &BTD::hci_connect;
  TRACE_USBHOST_HCI(printf("%lu - Executed connect. [ 0x%08.8X ]\r\n", ticks, hci_last_invoked);)
}/*}}}*/

void BTD::hci_link_key_notification() {/*{{{*/

  // See Core V4.0 Vol. 2.E.7.7.24, "Link Key Notification Event"
  // Event EV_LINK_KEY_NOTIFICATION

  // Perform HCI_Write_Stored_Link_Key command
  hci_write_stored_link_key();

  TRACE_USBHOST_HCI(printf("%lu - Executed link_key_notification. [ 0x%08.8X ]\r\n", ticks, hci_last_invoked);)

}/*}}}*/

void BTD::hci_delete_stored_link_key( uint8_t * bd_addr ) {/*{{{*/

  hci_event_flag &= ~HCI_FLAG_CMD_COMPLETE;
  hci_state = HCI_DELETE_STORED_LINK_KEYS;

  hcicmd[0] = 0x12;
  hcicmd[1] = 0x03 << 2;
  hcicmd[2] = 7;
  hcicmd[9] = 0x01;

  memset( &(hcicmd[3]), 0, 6 );
  if ( bd_addr ) memmove(&(hcicmd[3]), bd_addr, 6);

  HCI_Command(hcicmd, 10);
  hci_last_invoked = NULL;
  TRACE_USBHOST_HCI(printf("%lu - Executed delete_stored_link_key. [ 0x%08.8X ]\r\n", ticks, hci_last_invoked);)
}/*}}}*/

void BTD::hci_read_stored_link_key( uint8_t * bd_addr ) {/*{{{*/

  // 7.3.8 Read Stored Link Key Command
  hci_event_flag &= ~HCI_FLAG_CMD_COMPLETE;
  hci_state = (bd_addr == NULL) ? HCI_READ_STORED_LINK_KEYS : HCI_READ_STORED_LINK_KEY;

  hcicmd[0] = 0x0D;
  hcicmd[1] = 0x03 << 2;
  hcicmd[2] = 7; // Parameter length 
  memset( &(hcicmd[3]), 0, 7 );
  if ( bd_addr == NULL ) hcicmd[9] = 0x01; // Return all link keys
  else memmove( &(hcicmd[3]), bd_addr, 6 );

  // May generate zero or more Return Link Keys events

  HCI_Command(hcicmd, 10);
  hci_last_invoked = NULL;
  TRACE_USBHOST_HCI(printf("%lu - Executed read_stored_link_key. [ 0x%08.8X ]\r\n", ticks, hci_last_invoked);)

}/*}}}*/

void BTD::hci_write_stored_link_key() {/*{{{*/

  // 7.3.9 Write Stored Link Key Command 
  // Also see Core V4.0 Vol. 2.E.7.7.24, "Link Key Notification Event"
  // Event EV_LINK_KEY_NOTIFICATION

  // Perform HCI_Write_Stored_Link_Key command

  hci_event_flag &= ~HCI_FLAG_CMD_COMPLETE;
  hci_state = HCI_WRITE_STORED_LINK_KEY;

  hcicmd[0] = 0x11; // OCF = 0x11, 
  hcicmd[1] = 0x03 << 2; // OGF = 3, See 2.E.7.3.9
  hcicmd[2] = 23; // Parameter length 
  hcicmd[3] = 1; // Link keys to write
  memmove( &(hcicmd[4]), disc_bdaddr, 6 );
  memmove( &(hcicmd[10]), link_key, 16 );

  HCI_Command(hcicmd, 26);
  hci_last_invoked = NULL;
  TRACE_USBHOST_HCI(printf("%lu - Executed write_stored_link_key. [ 0x%08.8X ]\r\n", ticks, hci_last_invoked);)

}/*}}}*/

void BTD::hci_master_link_key() {/*{{{*/

  // See 7.1.18 Master Link Key Command
  // Generates Encryption Change event on completion.

  hcicmd[0] = 0x17;
  hcicmd[1] = 0x01 << 2;
  hcicmd[2] = 1; // Parameter length 
  hcicmd[3] = 0x00; // 0x01 - Use semi-permanent link keys; 0x00 - Use temporary link keys 

  HCI_Command(hcicmd, 4);
  hci_last_invoked = &BTD::hci_master_link_key;
  TRACE_USBHOST_HCI(printf("%lu - Executed master_link_key. [ 0x%08.8X ]\r\n", ticks, hci_last_invoked);)

}/*}}}*/

void BTD::hci_change_connection_link_key() {/*{{{*/

  // 7.1.17 Change Connection Link Key Command
  // Generates Encryption Change event on completion.

  hcicmd[0] = 0x15;
  hcicmd[1] = 0x01 << 2;
  hcicmd[2] = 2; // Parameter length 
  hcicmd[3] = (uint8_t)(hci_handle & 0xFF); //connection handle - low byte
  hcicmd[4] = (uint8_t)((hci_handle >> 8) & 0x0F); //connection handle - high byte

  HCI_Command(hcicmd, 5);
  hci_last_invoked = &BTD::hci_change_connection_link_key;
  TRACE_USBHOST_HCI(printf("%lu - Executed change_connection_link_key. [ 0x%08.8X ]\r\n", ticks, hci_last_invoked);)

}/*}}}*/

void BTD::hci_set_connection_encryption() {/*{{{*/

  // 7.1.16 Set Connection Encryption Command
  // Generates Encryption Change event on completion.

  hci_event_flag &= ~HCI_FLAG_ENCRYPTION_CHANGE_EVENT; 
  hci_state = HCI_ENCRYPTION_CHANGE;

  hcicmd[0] = 0x13;
  hcicmd[1] = 0x01 << 2;
  hcicmd[2] = 3; // Parameter length 
  hcicmd[3] = (uint8_t)(hci_handle & 0xFF); //connection handle - low byte
  hcicmd[4] = (uint8_t)((hci_handle >> 8) & 0x0F); //connection handle - high byte
  hcicmd[5] = 0x01; // Boolean, 0 - off; 1 - On. 

  HCI_Command(hcicmd, 6);
  hci_last_invoked = &BTD::hci_set_connection_encryption;
  TRACE_USBHOST_HCI(printf("%lu - Executed set_connection_encryption. [ 0x%08.8X ]\r\n", ticks, hci_last_invoked);)

}/*}}}*/

void BTD::hci_read_remote_supported_features() {/*{{{*/

  // 7.1.21 Read Remote Supported Features Command
  hci_event_flag &= ~HCI_FLAG_READ_REMOTE_SUPPORTED_FEATURES_EVENT;
  hci_state = HCI_READ_REMOTE_SUPPORTED_FEATURES;

  hcicmd[0] = 0x1B;
  hcicmd[1] = 0x01 << 2;
  hcicmd[2] = 2; // Parameter length 
  hcicmd[3] = (uint8_t)(hci_handle & 0xFF); //connection handle - low byte
  hcicmd[4] = (uint8_t)((hci_handle >> 8) & 0x0F); //connection handle - high byte

  HCI_Command(hcicmd, 5);
  hci_last_invoked = &BTD::hci_read_remote_supported_features;
  TRACE_USBHOST_HCI(printf("%lu - Executed read_remote_supported_features. [ 0x%08.8X ]\r\n", ticks, hci_last_invoked);)

}/*}}}*/

void BTD::hci_read_remote_extended_features() {/*{{{*/

  // 7.1.22 Read Remote Extended Features Command
  hci_event_flag &= ~HCI_FLAG_READ_REMOTE_EXTENDED_FEATURES_EVENT;
  hci_state = HCI_READ_REMOTE_EXTENDED_FEATURES;

  hcicmd[0] = 0x1C;
  hcicmd[1] = 0x01 << 2;
  hcicmd[2] = 2; // Parameter length 
  hcicmd[3] = (uint8_t)(hci_handle & 0xFF); //connection handle - low byte
  hcicmd[4] = (uint8_t)((hci_handle >> 8) & 0x0F); //connection handle - high byte
  hcicmd[5] = 0x01; // 0x00 - Normal LMP features, 0x01-0xFF - The corresponding page, returning zero if page not supported

  HCI_Command(hcicmd, 5);
  hci_last_invoked = &BTD::hci_read_remote_extended_features;
  TRACE_USBHOST_HCI(printf("%lu - Executed read_remote_extended_features. [ 0x%08.8X ]\r\n", ticks, hci_last_invoked);)

}/*}}}*/

void BTD::hci_read_remote_version_information() {/*{{{*/

  // 7.1.23 Read Remote Version Information Command
  hci_event_flag &= ~HCI_FLAG_READ_REMOTE_VERSION_INFORMATION_EVENT;
  hci_state = HCI_READ_REMOTE_VERSION_INFORMATION;

  hcicmd[0] = 0x1D;
  hcicmd[1] = 0x01 << 2;
  hcicmd[2] = 2; // Parameter length 
  hcicmd[3] = (uint8_t)(hci_handle & 0xFF); //connection handle - low byte
  hcicmd[4] = (uint8_t)((hci_handle >> 8) & 0x0F); //connection handle - high byte

  HCI_Command(hcicmd, 5);
  hci_last_invoked = &BTD::hci_read_remote_version_information;
  TRACE_USBHOST_HCI(printf("%lu - Executed read_remote_version_information. [ 0x%08.8X ]\r\n", ticks, hci_last_invoked);)

}/*}}}*/

void BTD::hci_pin_code_request_reply() {/*{{{*/

  hci_state = HCI_CONNECT_STATE; 

  if (0 == strlen(btdPin)) {
    TRACE_USBHOST(printf("No PIN configured; sending negative reply\r\n");)
    hci_pin_code_negative_request_reply();
    return;
  } 

  uint32_t cmdlen = 10;

  hcicmd[0] = 0x0D; // HCI OCF = 0D
  hcicmd[1] = 0x01 << 2; // HCI OGF = 1
  hcicmd[2] = 0x17; // parameter length 23
  memmove( &(hcicmd[3]), disc_bdaddr, 6 );
  hcicmd[9] = strlen(const_cast<char *>(&btdPin[0])); // Length of pin

  uint8_t i;
  for (i = 0; i < hcicmd[9]; i++) { // The maximum size of the pin is 16
    hcicmd[i + 10] = btdPin[i];
    cmdlen++;
  }
  for (; i < 16; i++) {
    hcicmd[i + 10] = 0x00; // The rest should be 0
    cmdlen++;
  }

  HCI_Command(hcicmd, cmdlen);
  hci_last_invoked = &BTD::hci_pin_code_request_reply;
  TRACE_USBHOST_HCI(printf("%lu - Executed pin_code_request_reply, total length %u. [ 0x%08.8X ]\r\n", ticks, cmdlen, hci_last_invoked);)
}/*}}}*/

void BTD::hci_pin_code_negative_request_reply() {/*{{{*/
  hcicmd[0] = 0x0E; // HCI OCF = 0E
  hcicmd[1] = 0x01 << 2; // HCI OGF = 1
  hcicmd[2] = 0x06; // parameter length 6
  memmove( &(hcicmd[3]), disc_bdaddr, 6 );

  HCI_Command(hcicmd, 9);
  hci_last_invoked = &BTD::hci_pin_code_negative_request_reply;
  TRACE_USBHOST_HCI(printf("%lu - Executed pin_code_negative_request_reply. [ 0x%08.8X ]\r\n", ticks, hci_last_invoked);)
}/*}}}*/

void BTD::hci_change_connection_packet_type() {/*{{{*/

  hci_event_flag &= ~(HCI_FLAG_CHANGE_CONNECTION_PACKET_TYPE | HCI_FLAG_CMD_COMPLETE);
  hci_state = HCI_CHANGE_CONNECTION_PACKET_TYPE;

  hcicmd[0] = 0x0F; // OCF
  hcicmd[1] = 0x01 << 2; // OGF
  hcicmd[2] = 0x03; // parameter length 6
  hcicmd[3] = (uint8_t)(hci_handle & 0xFF); //connection handle - low byte
  hcicmd[4] = (uint8_t)((hci_handle >> 8) & 0x0F); //connection handle - high byte
  // Bitmask 0xCC18 for ACL Link Type
  hcicmd[5] = 0x18;
  hcicmd[6] = 0xCC; 

  HCI_Command(hcicmd, 7);
  hci_last_invoked = &BTD::hci_change_connection_packet_type;
  TRACE_USBHOST_HCI(printf("%lu - Executed change_connection_packet_type. [ 0x%08.8X ]\r\n", ticks, hci_last_invoked);)
}/*}}}*/

void BTD::hci_link_key_request_reply() {/*{{{*/

  hci_state = HCI_CONNECT_STATE; // FIXME: Enter holding state

  hcicmd[0] = 0x0B; // HCI OCF = 0B
  hcicmd[1] = 0x01 << 2; // HCI OGF = 1
  hcicmd[2] = 0x06; // parameter length 6
  memmove( &(hcicmd[3]), disc_bdaddr, 6 );

  uint32_t cmdlen = 9;

  // FIXME: Understand what this link_key does :)
  // The link_key_generator string is sent back to the
  // connecting entity to serve as the link key.
  // I simply XOR in bits from the remote host BD_ADDR
  // to generate a unique link key for that host as
  // seen by this Arduino Due.

	for ( uint32_t i = 0 ; i < sizeof(link_key_generator) ; i++ ) {
		hcicmd[i+9] = pgm_read_byte(&link_key_generator[i]) ^ disc_bdaddr[i % 6];
    cmdlen++;
	}

  HCI_Command(hcicmd, cmdlen);
  hci_last_invoked = &BTD::hci_link_key_request_reply;
  TRACE_USBHOST_HCI(printf("%lu - Executed link_key_request_reply. [ 0x%08.8X ]\r\n", ticks, hci_last_invoked);)
}/*}}}*/

void BTD::hci_link_key_request_negative_reply() {/*{{{*/

  hci_state = HCI_CONNECT_STATE;

  hcicmd[0] = 0x0C; // HCI OCF = 0C
  hcicmd[1] = 0x01 << 2; // HCI OGF = 1
  hcicmd[2] = 0x06; // parameter length 6
  memmove( &(hcicmd[3]), disc_bdaddr, 6 );

  HCI_Command(hcicmd, 9);
  hci_last_invoked = &BTD::hci_link_key_request_negative_reply;
  TRACE_USBHOST_HCI(printf("%lu - Executed link_key_request_negative_reply. [ 0x%08.8X ]\r\n", ticks, hci_last_invoked);)
}/*}}}*/

void BTD::hci_write_authentication_enable(bool v) {/*{{{*/

  hci_event_flag &= ~HCI_FLAG_CMD_COMPLETE;
  hci_state = HCI_WRITE_AUTHENTICATION_ENABLE;

  hcicmd[0] = 0x20; // HCI OCF = 20 
  hcicmd[1] = 0x03 << 2; // HCI OGF = 1
  hcicmd[2] = 0x01; // parameter length = 1
  hcicmd[3] = v ? 0x01 : 0x00; 

  HCI_Command(hcicmd, 4);
  hci_last_invoked = NULL;
  TRACE_USBHOST_HCI(printf("%lu - Set Require Authentication Enable flag %s.\r\n", ticks, v ? "true" : "false");)
}/*}}}*/

void BTD::hci_authentication_request() {/*{{{*/
  hcicmd[0] = 0x11; // HCI OCF = 11
  hcicmd[1] = 0x01 << 2; // HCI OGF = 1
  hcicmd[2] = 0x02; // parameter length = 2
  hcicmd[3] = (uint8_t)(hci_handle & 0xFF); //connection handle - low byte
  hcicmd[4] = (uint8_t)((hci_handle >> 8) & 0x0F); //connection handle - high byte

  HCI_Command(hcicmd, 5);
  hci_last_invoked = &BTD::hci_authentication_request;
  TRACE_USBHOST_HCI(printf("%lu - Accept authentication request, hci_handle 0x%04.4X.\r\n", ticks, hci_handle);)
}/*}}}*/

void BTD::hci_disconnect(uint16_t handle) {/*{{{*/
  // This is called by the different services
  hci_event_flag &= ~HCI_FLAG_DISCONN_COMPLETE;
  hcicmd[0] = 0x06; // HCI OCF = 6
  hcicmd[1] = 0x01 << 2; // HCI OGF = 1
  hcicmd[2] = 0x03; // parameter length = 3
  hcicmd[3] = (uint8_t)(handle & 0xFF); //connection handle - low byte
  hcicmd[4] = (uint8_t)((handle >> 8) & 0x0F); //connection handle - high byte
  hcicmd[5] = 0x13; // reason

  HCI_Command(hcicmd, 6);
  hci_last_invoked = NULL;
  TRACE_USBHOST_HCI(printf("%lu - Executed disconnect. [ 0x%08.8X ]\r\n", ticks, hci_last_invoked);)
}/*}}}*/

void BTD::hci_write_class_of_device() {/*{{{*/
  // See http://bluetooth-pentest.narod.ru/software/bluetooth_class_of_device-service_generator.html
  hci_state = HCI_CLASS_STATE;
  hcicmd[0] = 0x24; // HCI OCF = 0x24 
  hcicmd[1] = 0x03 << 2; // HCI OGF = 3
  hcicmd[2] = 0x03; // parameter length = 3
  hcicmd[3] = 0x14; // Palm-sized PC/PDA 
  hcicmd[4] = 0x01; // Computer
  hcicmd[5] = 0xC2; // Networking, telephony, information

  HCI_Command(hcicmd, 6);
  hci_last_invoked = &BTD::hci_write_class_of_device;
  TRACE_USBHOST_HCI(printf("%lu - Executed write_class_of_device. [ 0x%08.8X ]\r\n", ticks, hci_last_invoked);)
}/*}}}*/

void BTD::hci_io_capability_response() {/*{{{*/

  // See Core V4.0 Vol. 2.E.7.1.29, "IO Capability Request Reply Command"
  // A call to hci_read_local_oob_data() must precede this call
  hci_event_flag &= ~HCI_FLAG_CMD_COMPLETE;
  hci_state = HCI_IO_CAPABILITY_RESPONSE_SENT_STATE;

  hcicmd[0] = 0x2B; // HCI OCF = 0x2B, See Core Spec Vol. 4 Part 2.E.7.1.29
  hcicmd[1] = 0x01 << 2; // HCI OGF = 1
  hcicmd[2] = 9; // parameter length
  memmove( &(hcicmd[3]), disc_bdaddr, 6); 
  hcicmd[9] = 0x03; // IO_Capability: 0x01 - DisplayYesNo 0x03 - NoInputNoOutput; 0x00 - DisplayOnly
  hcicmd[10] = /*bSspCapable ? 0x01 :*/ 0x00; // OOB_Data_Present: 0x01 - OOB authentication data present 
  hcicmd[11] = 0x05; // Authentication_Requirements: 0x05 - MITM Protection Required – General Bonding 

  HCI_Command(hcicmd, 12);
  hci_last_invoked = &BTD::hci_io_capability_response;
  TRACE_USBHOST_HCI(printf("%lu - Executed io_capability_response. [ 0x%08.8X ]\r\n", ticks, hci_last_invoked);)

}/*}}}*/

void BTD::hci_write_pin_type(bool variable) {/*{{{*/

  // See Core V4.0 Vol. 2.E.7.3.6, "Write PIN Type Command"
  hci_event_flag &= ~HCI_FLAG_CMD_COMPLETE;
  hci_state = HCI_WRITE_PIN_TYPE;

  hcicmd[0] = 0x0A; // OCF
  hcicmd[1] = 0x03 << 2; // HCI OGF = 1
  hcicmd[2] = 1; // parameter length
  hcicmd[3] = variable ? 0x00 : 0x01;

  HCI_Command(hcicmd, 4);
  hci_last_invoked = NULL; 
  TRACE_USBHOST_HCI(printf("%lu - Executed write_pin_type. [ 0x%08.8X ]\r\n", ticks, hci_last_invoked);)

}/*}}}*/

void BTD::hci_read_local_oob_data() {/*{{{*/

  hci_event_flag &= ~HCI_FLAG_CMD_COMPLETE;
  hci_state = HCI_READ_LOCAL_OOB_DATA_STATE;

  hcicmd[0] = 0x57; // HCI OCF = 0x57, See Core Spec Vol. 4 Part 2.E.7.3.60
  hcicmd[1] = 0x03 << 2; // HCI OGF = 3
  hcicmd[2] = 0; // parameter length

  HCI_Command(hcicmd, 3);
  hci_last_invoked = &BTD::hci_read_local_oob_data;
  TRACE_USBHOST_HCI(printf("%lu - Executed read_local_oob_data. [ 0x%08.8X ]\r\n", ticks, hci_last_invoked);)

}/*}}}*/

void BTD::hci_oob_data_request_reply() {/*{{{*/

  // Executed if param 10 (OOB Data Present) set to 0x01 in 
  // BTD::oob_data_request_reply()

  hci_event_flag &= ~HCI_FLAG_CMD_COMPLETE;
  hci_state = HCI_OOB_DATA_REQUEST_REPLY_SENT_STATE;

  memset(hcicmd, 0, sizeof(hcicmd));

  hcicmd[0] = 0x30; // HCI OCF = 0x57, See Core Spec Vol. 4 Part 2.E.7.3.60
  hcicmd[1] = 0x01 << 2; // HCI OGF = 3
  hcicmd[2] = 38; // parameter length
  memmove( &(hcicmd[3]), disc_bdaddr, 6); 
  memmove( &(hcicmd[9]), oob_pairing_hash, 16);
  memmove( &(hcicmd[25]), oob_pairing_randomizer, 16);

  TRACE_USBHOST_HCI(printf("%lu - Executed oob_data_request_reply. [ 0x%08.8X ]\r\n", ticks, hci_last_invoked);)
  hci_dump_command();

  HCI_Command(hcicmd, 41);
  hci_last_invoked = &BTD::hci_oob_data_request_reply;

}/*}}}*/


/*******************************************************************
 *                                                                 *
 *                        HCI ACL Data Packet                      *
 *                                                                 *
 *   buf[0]          buf[1]          buf[2]          buf[3]
 *   0       4       8    11 12      16              24            31 MSB
 *  .-+-+-+-+-+-+-+-|-+-+-+-|-+-|-+-|-+-+-+-+-+-+-+-|-+-+-+-+-+-+-+-.
 *  |      HCI Handle       |PB |BC |       Data Total Length       |   HCI ACL Data Packet
 *  .-+-+-+-+-+-+-+-|-+-+-+-|-+-|-+-|-+-+-+-+-+-+-+-|-+-+-+-+-+-+-+-.
 *
 *   buf[4]          buf[5]          buf[6]          buf[7]
 *   0               8               16                            31 MSB
 *  .-+-+-+-+-+-+-+-|-+-+-+-+-+-+-+-|-+-+-+-+-+-+-+-|-+-+-+-+-+-+-+-.
 *  |            Length             |          Channel ID           |   Basic L2CAP header
 *  .-+-+-+-+-+-+-+-|-+-+-+-+-+-+-+-|-+-+-+-+-+-+-+-|-+-+-+-+-+-+-+-.
 *
 *   buf[8]          buf[9]          buf[10]         buf[11]
 *   0               8               16                            31 MSB
 *  .-+-+-+-+-+-+-+-|-+-+-+-+-+-+-+-|-+-+-+-+-+-+-+-|-+-+-+-+-+-+-+-.
 *  |     Code      |  Identifier   |            Length             |   Control frame (C-frame)
 *  .-+-+-+-+-+-+-+-|-+-+-+-+-+-+-+-|-+-+-+-+-+-+-+-|-+-+-+-+-+-+-+-.   (signaling packet format)
 */

/************************************************************/
/*                    L2CAP Commands                        */
/************************************************************/

void RFCOMM_PROTOCOL_MESSAGE_handler(usb_add_t add,
    usb_ep_t ep,
    uhd_trans_status_t status,
    iram_size_t nb_transferred,
    uint8_t callback_data)
{/*{{{*/
  btd_instance->invoke_callback_state_update(add, ep, status, nb_transferred, callback_data);
}/*}}}*/

void RFCOMM_PROTOCOL_CREDIT_handler(usb_add_t add,
    usb_ep_t ep,
    uhd_trans_status_t status,
    iram_size_t nb_transferred,
    uint8_t callback_data)
{/*{{{*/
  btd_instance->invoke_callback_state_update(add, ep, status, nb_transferred, callback_data);
}/*}}}*/

void L2CAP_Command_handler( usb_add_t add,
    usb_ep_t ep,
    uhd_trans_status_t status,
    iram_size_t nb_transferred,
    uint8_t callback_data )
{/*{{{*/
  btd_instance->invoke_callback_state_update(add, ep, status, nb_transferred, callback_data);
}/*}}}*/

bool BTD::L2CAP_Command(uint16_t handle, uint8_t* data, uint8_t nbytes, uint8_t channelLow, uint8_t channelHigh, uhd_callback_trans_t callback, uint8_t callback_data) {/*{{{*/

  irqflags_t flags = cpu_irq_save();
  uint32_t transfer_length = (8 + nbytes);
  uint32_t rcode; 

  memmove( l2capoutbuf + 8, l2capoutbuf, nbytes );

  // HCI handle with PB,BC flag
  l2capoutbuf[0] = (uint8_t)(handle & 0xff);
  l2capoutbuf[1] = (uint8_t)(((handle >> 8) & 0x0f) | 0x20);
  // HCI ACL total data length
  l2capoutbuf[2] = (uint8_t)((4 + nbytes) & 0xff);
  l2capoutbuf[3] = (uint8_t)((4 + nbytes) >> 8);
  l2capoutbuf[4] = (uint8_t)(nbytes & 0xff); // L2CAP header: Length
  l2capoutbuf[5] = (uint8_t)(nbytes >> 8);
  l2capoutbuf[6] = channelLow;
  l2capoutbuf[7] = channelHigh;

  // If the queue index is zero, do not issue the request - only dump it.
  if ( callback_data == 0 ) {
    printf("~~ Data for handle 0x%04.4X, length %lu", handle, transfer_length);
    for (uint16_t i = 0; i < L2CAP_OUT_BUFFERSIZE; i++) {
      if (!(i % 16)) printf("\r\n%lu ", GetTickCount());
      if (!(i % 8)) printf(" ");
      printf(" %02.2x", l2capoutbuf[i]);
    }
    printf("\r\n");
  }
  else if (!uhd_ep_run(
    bAddress,
    epInfo[BTD_DATAOUT_PIPE].deviceEpNum,
    true, 
    this->l2capoutbuf,
    transfer_length,
    20,
    callback == NULL ? &L2CAP_Command_handler : callback,
    callback_data // Should contain a callback queue entry number + 1
  )) {
    printf("~~ Error sending L2CAP message of size %u bytes, Channel ID: 0x%02.2x %02.2x\r\n", nbytes, channelHigh, channelLow);
    cpu_irq_restore(flags);
    return false;
  }
  cpu_irq_restore(flags);
  return true;
}/*}}}*/

void BTD::l2cap_connection_request(uint16_t handle, uint8_t rxid, uint8_t* scid, uint16_t psm) {/*{{{*/

  l2capoutbuf[0] = L2CAP_CMD_CONNECTION_REQUEST; // Code
  l2capoutbuf[1] = rxid; // Identifier
  l2capoutbuf[2] = 0x04; // Length
  l2capoutbuf[3] = 0x00;
  l2capoutbuf[4] = (uint8_t)(psm & 0xff); // PSM
  l2capoutbuf[5] = (uint8_t)(psm >> 8);
  l2capoutbuf[6] = scid[0]; // Source CID
  l2capoutbuf[7] = scid[1];

  L2CAP_Command(handle, l2capoutbuf, 8);
}/*}}}*/

void BTD::l2cap_disconnection_request(uint16_t handle, uint8_t rxid, uint8_t* dcid, uint8_t* scid) {/*{{{*/
  l2capoutbuf[0] = L2CAP_CMD_DISCONNECT_REQUEST; // Code
  l2capoutbuf[1] = rxid; // Identifier
  l2capoutbuf[2] = 0x04; // Length
  l2capoutbuf[3] = 0x00;
  l2capoutbuf[4] = dcid[0];
  l2capoutbuf[5] = dcid[1];
  l2capoutbuf[6] = scid[0];
  l2capoutbuf[7] = scid[1];

  L2CAP_Command(handle, l2capoutbuf, 8);
}/*}}}*/

bool BTD::enqueue_l2cap_message(uint8_t bt_service_index, l2cap_queue_item_t type, uint8_t *scid, uint8_t transactionIDLow, uint8_t transactionIDHigh, uint32_t callback_payload )
{/*{{{*/

  if ( transfer_queue_full() )
    return false;

  uhd_l2cap_queue_t & current = tq_entry(type);

  current.type                    = type;
  current.bt_service_index        = bt_service_index;
  current.t.sdp.payload           = callback_payload;
  current.handle                  = hci_handle;
  current.t.sdp.service_class_id  = scid[1] << 8 | scid[0];
  current.t.sdp.transactionIDLow  = transactionIDLow;
  current.t.sdp.transactionIDHigh = transactionIDHigh;

  current.l2capoutbuf[0] = SDP_SERVICE_SEARCH_ATTRIBUTE_RESPONSE_PDU;
  // SDP TransactionIDs uniquely identify request PDUs.
  // The TransactionID of a response is required to be identical to that
  // of the request that triggers it.
  current.l2capoutbuf[1] = transactionIDHigh;
  current.l2capoutbuf[2] = transactionIDLow;

  switch ( current.type ) {

    case L2CAP_SDP_SERVICE_ATTRIBUTE_RESPONSE_PBR:
      // Message must be prepared at point of dispatch
      // (exceeds size of queue element buffer)
      current.l2capoutbuf[3] = 0x00;
      current.l2capoutbuf[4] = 0x00; // Parameter Length
      current.l2capoutbuf[5] = 0x00;
      current.l2capoutbuf[6] = 0x00; // AttributeListsByteCount
      current.message_length = 7;
      break;

    case L2CAP_SERVICE_NOT_SUPPORTED:
      current.l2capoutbuf[3] = 0x00;
      current.l2capoutbuf[4] = 0x00; // Parameter Length
      current.l2capoutbuf[5] = 0x00;
      current.l2capoutbuf[6] = 0x00; // AttributeListsByteCount
      current.message_length = 7;
      break;

    default: 
      current.message_length = 3;
      break;
  }

  return true;

}/*}}}*/

bool BTD::enqueue_l2cap_message_initiator(uint8_t bt_service_index, l2cap_queue_item_t type, uint16_t handle, uint8_t rxid, uint8_t infoTypeLow, uint8_t infoTypeHigh )
{/*{{{*/

  if ( transfer_queue_full() )
    return false;

  uhd_l2cap_queue_t & current = tq_entry(type);

  current.type             = type;
  current.bt_service_index = bt_service_index;

  switch ( type ) {
    case L2CAP_QUEUE_INFORMATION_REQUEST:

      current.handle            = handle;
      current.t.rq.rxid         = rxid;
      current.t.rq.infoTypeLow  = infoTypeLow;
      current.t.rq.infoTypeHigh = infoTypeHigh;

      current.l2capoutbuf[0] = L2CAP_CMD_INFORMATION_REQUEST; // Code
      current.l2capoutbuf[1] = rxid; // Identifier
      current.l2capoutbuf[2] = 0x02; // Length
      current.l2capoutbuf[3] = 0x00;
      current.l2capoutbuf[4] = infoTypeLow;
      current.l2capoutbuf[5] = infoTypeHigh;
      current.message_length = 6;

      break;

    case L2CAP_QUEUE_INFORMATION_RESPONSE:

      current.handle            = handle;
      current.t.re.rxid         = rxid;
      current.t.re.infoTypeLow  = infoTypeLow;
      current.t.re.infoTypeHigh = infoTypeHigh;

      current.l2capoutbuf[0] = L2CAP_CMD_INFORMATION_RESPONSE; // Code
      current.l2capoutbuf[1] = rxid; // Identifier
      current.l2capoutbuf[2] = 0x08; // Length
      current.l2capoutbuf[3] = 0x00;
      current.l2capoutbuf[4] = infoTypeLow;
      current.l2capoutbuf[5] = infoTypeHigh;
      current.l2capoutbuf[6] = 0x00; // Result = success
      current.l2capoutbuf[7] = 0x00; // Result = success
      current.l2capoutbuf[8] = 0x00;
      current.l2capoutbuf[9] = 0x00;
      current.l2capoutbuf[10] = 0x00;
      current.l2capoutbuf[11] = 0x00;
      current.message_length = 12;

      break;

    default:
      return false;
      break;
  }
  return true;
}/*}}}*/

bool BTD::enqueue_l2cap_connection_response(uint8_t bt_service_index, uint16_t handle, uint8_t rxid, uint8_t* dcid, uint8_t* scid, uint8_t result, uint32_t callback_payload)
{/*{{{*/
  if ( transfer_queue_full() )
    return false;

  uhd_l2cap_queue_t & current = tq_entry(L2CAP_CONNECTION_RESPONSE);

  current.type                = L2CAP_CONNECTION_RESPONSE;
  current.bt_service_index    = bt_service_index;
  current.t.conn_resp.payload = callback_payload;
  current.handle              = handle;
  current.t.conn_resp.rxid    = rxid;
  current.t.conn_resp.dcid    = dcid[1] << 8 | dcid[0];
  current.t.conn_resp.scid    = scid[1] << 8 | scid[0];
  current.t.conn_resp.result  = result;

  current.l2capoutbuf[0] = L2CAP_CMD_CONNECTION_RESPONSE; // Code
  current.l2capoutbuf[1] = rxid; // Identifier
  current.l2capoutbuf[2] = 0x08; // Length
  current.l2capoutbuf[3] = 0x00;
  current.l2capoutbuf[4] = dcid[0]; // Destination CID
  current.l2capoutbuf[5] = dcid[1];
  current.l2capoutbuf[6] = scid[0]; // Source CID
  current.l2capoutbuf[7] = scid[1];
  current.l2capoutbuf[8] = result; // Result: Pending or Success
  current.l2capoutbuf[9] = 0x00;
  current.l2capoutbuf[10] = 0x00; // No further information
  current.l2capoutbuf[11] = 0x00;
  current.message_length = 12;

  return true;
}/*}}}*/

bool BTD::enqueue_l2cap_config_request(uint8_t bt_service_index, uint16_t handle, uint8_t rxid, uint8_t* scid, uint32_t callback_payload)
{/*{{{*/

  if ( transfer_queue_full() )

    return false;

  uhd_l2cap_queue_t & current = tq_entry(L2CAP_CONFIG_REQUEST);

  current.type               = L2CAP_CONFIG_REQUEST;
  current.bt_service_index   = bt_service_index;
  current.t.conf_req.payload = callback_payload;

  current.handle             = handle;
  current.t.conf_req.rxid    = rxid;
  current.t.conf_req.scid    = scid[1] << 8 | scid[0];

  current.l2capoutbuf[0] = L2CAP_CMD_CONFIG_REQUEST; // Code
  current.l2capoutbuf[1] = rxid; // Identifier
  current.l2capoutbuf[2] = 0x04; // Length
  current.l2capoutbuf[3] = 0x00;
  current.l2capoutbuf[4] = scid[0]; // Destination CID
  current.l2capoutbuf[5] = scid[1];
  current.l2capoutbuf[6] = 0x00; // Flags
  current.l2capoutbuf[7] = 0x00;
  current.message_length = 8;
#if 0
  current.l2capoutbuf[8] = 0x01; // Config Opt: type = MTU (Maximum Transmission Unit) - Hint
  current.l2capoutbuf[9] = 0x02; // Config Opt: length
  current.l2capoutbuf[10] = 0xFF; // MTU
  current.l2capoutbuf[11] = 0xFF;

  current.message_length = 12;
#endif

  return true;
}/*}}}*/

bool BTD::enqueue_l2cap_config_response(uint8_t bt_service_index, uint16_t handle, uint8_t rxid, uint8_t* scid, uint32_t callback_payload)
{/*{{{*/
  if ( transfer_queue_full() )
    return false;

  uhd_l2cap_queue_t & current = tq_entry(L2CAP_CONFIG_RESPONSE);

  current.type               = L2CAP_CONFIG_RESPONSE;
  current.bt_service_index   = bt_service_index;
  current.t.conf_req.payload = callback_payload;
  current.handle             = handle;
  current.t.conf_req.rxid    = rxid;
  current.t.conf_req.scid    = scid[1] << 8 | scid[0];

  current.l2capoutbuf[0] = L2CAP_CMD_CONFIG_RESPONSE; // Code
  current.l2capoutbuf[1] = rxid; // Identifier
  current.l2capoutbuf[2] = 0x0A; // Length
  current.l2capoutbuf[3] = 0x00;
  current.l2capoutbuf[4] = scid[0]; // Destination CID
  current.l2capoutbuf[5] = scid[1];
  current.l2capoutbuf[6] = 0x00; // Flag
  current.l2capoutbuf[7] = 0x00;
  current.l2capoutbuf[8] = 0x00; // Result
  current.l2capoutbuf[9] = 0x00;
  current.l2capoutbuf[10] = 0x01; // Config - MTU (0x01)
  current.l2capoutbuf[11] = 0x02; // - Length 2 bytes
  current.l2capoutbuf[12] = 0xFF; // - MTU value 0xFFFF 
  current.l2capoutbuf[13] = 0xFF;

  current.message_length = 14;

  return true;
}/*}}}*/

bool BTD::enqueue_l2cap_disconnection_request(uint8_t bt_service_index, uint16_t handle, uint8_t rxid, uint8_t* dcid, uint8_t* scid, uint32_t callback_payload)
{/*{{{*/
  if ( transfer_queue_full() )
    return false;

  uhd_l2cap_queue_t & current = tq_entry(L2CAP_DISCONNECT_REQUEST);

  current.type             = L2CAP_DISCONNECT_REQUEST;
  current.bt_service_index = bt_service_index;
  current.t.disc.payload   = callback_payload;
  current.handle           = handle;
  current.t.disc.rxid      = rxid;
  current.t.disc.dcid      = dcid[1] << 8 | dcid[0];
  current.t.disc.scid      = scid[1] << 8 | scid[0];

  current.l2capoutbuf[0] = L2CAP_CMD_DISCONNECT_REQUEST; // Code
  current.l2capoutbuf[1] = rxid; // Identifier
  current.l2capoutbuf[2] = 0x04; // Length
  current.l2capoutbuf[3] = 0x00;
  current.l2capoutbuf[4] = dcid[0];
  current.l2capoutbuf[5] = dcid[1];
  current.l2capoutbuf[6] = scid[0];
  current.l2capoutbuf[7] = scid[1];

  current.message_length = 8;

  return true;
}/*}}}*/

bool BTD::enqueue_l2cap_disconnection_response(uint8_t bt_service_index, uint16_t handle, uint8_t rxid, uint8_t* dcid, uint8_t* scid, uint32_t callback_payload)
{/*{{{*/
  if ( transfer_queue_full() )
    return false;

  uhd_l2cap_queue_t & current = tq_entry(L2CAP_DISCONNECT_RESP);

  current.type             = L2CAP_DISCONNECT_RESP;
  current.bt_service_index = bt_service_index;
  current.t.disc.payload   = callback_payload;

  current.handle           = handle;
  current.t.disc.rxid      = rxid;
  current.t.disc.dcid      = dcid[1] << 8 | dcid[0];
  current.t.disc.scid      = scid[1] << 8 | scid[0];

  current.l2capoutbuf[0] = L2CAP_CMD_DISCONNECT_RESPONSE;
  current.l2capoutbuf[1] = rxid; // Identifier
  current.l2capoutbuf[2] = 0x04; // Length
  current.l2capoutbuf[3] = 0x00;
  current.l2capoutbuf[4] = dcid[0];
  current.l2capoutbuf[5] = dcid[1];
  current.l2capoutbuf[6] = scid[0];
  current.l2capoutbuf[7] = scid[1];

  current.message_length = 8;

  return true;
}/*}}}*/

/************************************************************/
/*                    RFCOMM Commands                       */
/************************************************************/

void BTD::RFCOMM_Command(uint8_t *rfcomm_scid, uint8_t* data, uint8_t nbytes) {
	L2CAP_Command(hci_handle, data, nbytes, rfcomm_scid[0], rfcomm_scid[1]);
}

bool BTD::enqueue_sendRfcomm(
    uint8_t bt_service_index,
    uint16_t handle,
    uint8_t *rfcomm_scid,

    uint8_t channel,
    uint8_t direction,
    uint8_t CR,

    uint8_t channelType,
    uint8_t pfBit,
    uint8_t* data, uint8_t length,
    uint32_t callback_payload)
{/*{{{*/

  if ( transfer_queue_full() )
    return false;

  uhd_l2cap_queue_t & current = tq_entry();

  current.type              = RFCOMM_PROTOCOL_MESSAGE;
  current.handle            = handle;
  current.bt_service_index  = bt_service_index;
  current.t.rfcomm.payload  = callback_payload;
  current.t.rfcomm.scid     = rfcomm_scid[1] << 8 | rfcomm_scid[0];
  current.t.rfcomm.address  = channel | direction | CR | extendAddress; // RFCOMM Address
  current.t.rfcomm.control  = channelType | pfBit; // RFCOMM Control
  current.t.rfcomm.length   = length;
  memmove( &current.l2capoutbuf, data, length );

  return true;

}/*}}}*/

bool BTD::enqueue_sendRfcommCredit(uint8_t bt_service_index, uint16_t handle, uint8_t *rfcomm_scid, uint8_t channel, uint8_t direction, uint8_t CR, uint8_t channelType, uint8_t pfBit, uint8_t credit, uint32_t callback_payload)
{/*{{{*/

  if ( transfer_queue_full() )
    return false;

  uhd_l2cap_queue_t & current = tq_entry();

  current.type             = RFCOMM_PROTOCOL_CREDIT;
  current.handle           = handle;
  current.bt_service_index = bt_service_index;
  current.t.rfcomm.payload = callback_payload;
  current.t.rfcomm.scid    = rfcomm_scid[1] << 8 | rfcomm_scid[0];
  current.t.rfcomm.address = channel | direction | CR | extendAddress; // RFCOMM Address
  current.t.rfcomm.control = channelType | pfBit; // RFCOMM Control
  current.t.rfcomm.length  = 0;
  current.l2capoutbuf[0]   = credit;

  return true;

}/*}}}*/

/* CRC on 2 bytes */
uint8_t BTD::__crc(uint8_t* data) {/*{{{*/
	return (pgm_read_byte(&rfcomm_crc_table[pgm_read_byte(&rfcomm_crc_table[0xff ^ data[0]]) ^ data[1]]));
}/*}}}*/

uint8_t BTD::calcFcs(uint8_t *data) {/*{{{*/
  // Calculate FCS (frame check sequence)
  // See 5.2.1.6 of "Frame Checking Sequence Field (FCS)"
  // GSM 07.10 version 6.4.0 Release 1997
	if ((data[1] & 0xEF) == RFCOMM_UIH)
		return (0xff - __crc(data)); // FCS on 2 bytes
	else
		return (0xff - pgm_read_byte(&rfcomm_crc_table[__crc(data) ^ data[2]])); // FCS on 3 bytes
}/*}}}*/


/** Service Discovery Protocol parser */

#ifdef DEBUG_SDP_PDU
const char * sdp_pdu_type_descriptor_str[] {
  "NIL"    , // 0
  "UINT"   , // 1
  "INT"    , // 2
  "UUID"   , // 3
  "STR"    , // 4
  "BOOL"   , // 5
  "DATASEQ", // 6
  "DATAALT", // 7
  "URL"    , // 8
  "<CPC>"  
};
#endif

bool SdpParser::data_element_sequence_tokenizer( sdp_data_element_t *s,  uint8_t ** pdu_token ) {

  // Advance *pdu_token to the next data element
  // Return false if unable to do so (end of parse or invalid data element)
  uint32_t pdu_offset = 0;
  uint32_t data_length = 0;

  s->complete        = 0;
  s->fragment_length = 0;
  s->fragment        = &((*pdu_token)[pdu_offset]);
  s->type_descriptor = (*pdu_token)[pdu_offset];
  s->type_size       = s->type_descriptor & 0x07;
  s->type_descriptor >>= 3;

  if ( s->type_descriptor == SDP_PDU_CONTAINER_END ) {/*{{{*/
#ifdef DEBUG_SDP_PDU
    printf("> --------\r\n");
#endif
    s->parse_limit--;
  }/*}}}*/
  else if ( s->type_descriptor > SDP_PDU_MAX ) {/*{{{*/
    // Invalid type descriptor
    printf("!! Bad SDP PDU type desc %u\r\n", s->type_descriptor);
    return false;
  }/*}}}*/

  if ( s->parse_limit > 0 && s->working_offset >= s->parse_limit ) {
    return false;
  }

  // Invalid type size for the given descriptor
  if (!((0x01 << s->type_size) & pgm_read_byte(&sdp_pdu_size_mask[s->type_descriptor]))) { /*{{{*/
    printf("!! Invalid type size 0x%02.2X desc 0x%02.2X\r\n",
      s->type_size, 
      s->type_descriptor
    );
    return false;
  }/*}}}*/

  if ( s->type_size >= 5 ) {/*{{{*/

    // Type sizes 5, 6, and 7 indicate that the next 
    // 8, 16 or 32 bits (1, 2, or 4 bytes) specify the length
    // of the actual data element.
    //
    // When a type descriptor is received having data_length == 0,
    // we push the offset location and current size to the container_sizes
    // stack. That stack aggregates the sizes of all elements found at the
    // current structure nesting level. An SDP_CloseContainer() pseudo-element
    // in the parse stream causes the .size element value to be written into
    // the location .offset (base[offset] if base != NULL) at stack position 0.
    // This method sets the [complete] flag, updates working pointers, decrements
    // the parse depth index,  and immediately returns to the caller.
    // Caller then updates the output buffer at the indicated
    // position by writing the .size value into output buffer position .offset.
    //
    // Stack element 0 is used to return container sizes. 
    uint8_t skip_bytes = 1<<(s->type_size - 5);

    // Move to next byte
    pdu_offset++;

    // printf("\r\n> Length in next %u bytes\r\n", skip_bytes);
    switch ( s->type_size - 5 ) {
      case 0: s->data_length = *((uint8_t *)(&(*pdu_token)[pdu_offset])); break; 
      case 1: s->data_length = byteflip(*((uint16_t *)(&(*pdu_token)[pdu_offset]))); break; 
      case 2: 
        s->data_length = byteflip(*((uint16_t *)(&(*pdu_token)[pdu_offset+2]))) << 16;
        s->data_length |= byteflip(*((uint16_t *)(&(*pdu_token)[pdu_offset])));
        break; 
    }

    // Incompletely-specified data element may be a string, URL, or DE sequence
    switch ( s->type_descriptor ) {/*{{{*/
      case SDP_PDU_DATASEQ: // = 6,
      case SDP_PDU_DATAALT: // = 7,
        s->fragment_length = 1 + skip_bytes;
        if ( s->data_length == 0 ) {
          s->parse_depth++;
          // The working offset from a previous run
          // points to the start of this DE (the type descriptor).
          // This allows us to copy the computed length using the correct
          // data type size (uint8_t, uint16_t, or uint32_t, as above). 
          s->container_sizes[s->parse_depth].offset = s->working_offset;
          s->container_sizes[s->parse_depth].size = 0; 
#ifdef DEBUG_SDP_PDU
          printf("\r\n++ Stack %u @ 0x%04.4X", s->parse_depth, s->working_offset); 
          s->dump_stack();
#endif
          s->complete = 0;
        }
        // Working offset for the next element is simply the current element's
        // offset + number of size specifier bytes
        s->working_offset += s->fragment_length; 
        for ( uint16_t stack_index = s->parse_depth ; stack_index > 0 ; stack_index-- ) {
          s->container_sizes[stack_index].size += s->fragment_length; 
        }
        break;
      case SDP_PDU_STR: // = 4,
      case SDP_PDU_URL: // = 8,
        s->fragment_length = 1 + skip_bytes + s->data_length;
        s->working_offset += s->fragment_length;
      default:
        if ( s->data_length == 0 ) {/*{{{*/
          // We have an invalid PDU type descriptor. String and URL lengths MUST
          // be specified in code before this method is used. 
          printf("!! Invalid PDU type descriptor. String length must be specified @ %u\r\n",
              s->working_offset );
          return false;
        }/*}}}*/
        for ( uint16_t stack_index = s->parse_depth ; stack_index > 0 ; stack_index-- ) {
          s->container_sizes[stack_index].size += s->fragment_length; 
        }
        break;
    }/*}}}*/

    // Set pdu_offset to start of data element
    pdu_offset += skip_bytes; 

    // Data element begins right after the size specification bytes
    s->data = &((*pdu_token)[pdu_offset]); 

    // Fragment to dump is just the type descriptor and size spec bytes

  }/*}}}*/
  else {/*{{{*/

    // printf("\r\n> Length embedded\r\n");
    // Type sizes 0-4 correspond to a data length of 2 ^ (type_size)
    // bytes, following the (type descriptor | type_size) byte.
    // There is no separate data length word; data elements bearing
    // this type size specifier are always complete.
    // sdp_data_element_t should point to the data immediately following
    // the type descriptor byte.
    if ( s->type_descriptor == SDP_PDU_CONTAINER_END ) {
      s->data_length = 0;
    }
    else if ( s->type_descriptor == SDP_PDU_NIL ) {
      pdu_offset++;
      s->data_length = 1;
      s->fragment_length = 1;
    }
    else {
      pdu_offset++;
      s->data_length = 1<<(s->type_size);
      s->fragment_length = 1 + s->data_length;
    }
    s->data = &((*pdu_token)[pdu_offset]); 

    if ( s->type_descriptor != SDP_PDU_CONTAINER_END ) {
      s->working_offset += s->fragment_length;
      if ( s->parse_depth > 0 ) {
        for ( uint16_t stack_index = s->parse_depth ; stack_index > 0 ; stack_index-- ) {
          s->container_sizes[stack_index].size += s->fragment_length; 
        }
      }
    }


  }/*}}}*/
  
  // Set tokenization pointer to the first location inside 
  // a data element sequence, OR the start of the next data element 

  switch ( s->type_descriptor ) {/*{{{*/
    case SDP_PDU_NIL: // = 0,
    case SDP_PDU_UINT: // = 1,
    case SDP_PDU_INT: // = 2,
    case SDP_PDU_UUID: // = 3,
    case SDP_PDU_STR: // = 4,
    case SDP_PDU_BOOL: // = 5,
    case SDP_PDU_URL: // = 8,
      // Move tokenizer pointer past end of the current data element
      *pdu_token = &(s->data[s->data_length]);
      break;
    case SDP_PDU_DATASEQ: // = 6,
    case SDP_PDU_DATAALT: // = 7,
      // Move tokenizer to the first data element inside the container;
      // this enables us to tally up the lengths of child DEs.  
      *pdu_token = &(s->data[0]);
      break;
    case SDP_PDU_CONTAINER_END:
      // Simply advance the tokenizer position beyond the marker.
      // Copy the contents of the current stack entry into the bottom element,
      // and set the flag .complete, indicating that a container's size
      // specifier has been computed from it's contents. 
      *pdu_token = &((*pdu_token)[1]);
      s->data = &((*pdu_token)[0]);
      // Nothing to transfer to output
      s->data_length = 0;
      if ( s->parse_depth > 0 ) {
        s->container_sizes[0].size   = s->container_sizes[s->parse_depth].size;
        s->container_sizes[0].offset = s->container_sizes[s->parse_depth].offset;
        s->container_sizes[s->parse_depth].size   = 0;
        s->container_sizes[s->parse_depth].offset = 0;
        s->complete = 1;
        s->parse_depth--;
        // Propagate the DE sequence / alternative container size through each
        // successive stack layer? 
#ifdef DEBUG_SDP_PDU
        printf("\r\n0x%04.4X -- Pop stack: %u @ 0x%04.4X depth %u\r\n", 
          s->working_offset,
          s->container_sizes[0].size,
          s->container_sizes[0].offset,
          s->parse_depth
        );
        s->dump_stack();
#endif
      }
      break;
    default:
      // We probably have an invalid PDU
      return false;
      break;
  }/*}}}*/

#ifdef DEBUG_SDP_PDU

  printf("** %s[% 4.4u] @ %u, %u (desc 0x%02.2X:0x%02.2X - 0x%02.2X)",
    sdp_pdu_type_descriptor_str[s->type_descriptor],
    s->data_length,
    s->parse_depth,
    s->working_offset,
    s->type_descriptor,
    s->type_size, 
    (s->type_descriptor << 3) | s->type_size
  );
  // Data
  if ( s->type_descriptor < SDP_PDU_MAX )
  for ( pdu_offset = 0 ; pdu_offset < s->data_length ; pdu_offset++ ) {
    if (!(pdu_offset % 16)) printf("\r\n   ");
    if (!(pdu_offset % 8)) printf(" ");
    printf(" 0x%02.2x", s->data[pdu_offset]);
  } 
  // Lookahead
  for ( pdu_offset = 0 ; pdu_offset < 8 ; pdu_offset++ ) {
    if (!(pdu_offset % 16)) printf("\r\n%c  ", pdu_offset == 0 ? '>' : ' ');
    if (!(pdu_offset % 8)) printf(" ");
    printf(" 0x%02.2x", (*pdu_token)[pdu_offset]);
  } 
  printf("... \r\n");

#endif

  return true;
}

void sdp_data_element::dump_stack() {/*{{{*/
  printf("\r\n");
  for ( uint8_t i = 0 ; i < 8 ; i++ ) {
    printf("~~ %u -- %u @ 0x%04.4X\r\n", 
      i,
      container_sizes[i].size,
      container_sizes[i].offset
    );
  }
  printf("\r\n");
}/*}}}*/

bool sdp_data_element::update_typespec(uint8_t * pdu, size_offset_t * stack_entry) {/*{{{*/
  
  uint32_t size_value      = stack_entry->size;
  uint8_t type_descriptor_ = pdu[stack_entry->offset];
  uint8_t type_size_       = type_descriptor_ & 0x07;

  type_descriptor_ >>= 3;

  if ( type_size_ >= 5 && type_descriptor_ == SDP_PDU_DATASEQ ) {/*{{{*/

    // Type sizes 5, 6, and 7 indicate that the next 
    // 8, 16 or 32 bits (1, 2, or 4 bytes) specify the length
    // of the actual data element.
    uint8_t skip_bytes = 1<<(type_size_ - 5);

    // Move past type specification byte
    stack_entry->offset++;

    // Subtract size of header and skipped bytes from size value
    size_value -= (1 + skip_bytes);

    switch ( type_size_ - 5 ) {
      case 0: pdu[stack_entry->offset] = (uint8_t)(size_value & 0xFF); break; 
      case 1: *((uint16_t *)&(pdu[stack_entry->offset])) = (uint16_t)(size_value & 0xFFFF) ; break;
      case 2: *((uint32_t *)&(pdu[stack_entry->offset])) = size_value; break;
      default: return false;
    }

    TRACE_USBHOST(printf("\r\n+++  Update %u @ 0x%04.4X", size_value, stack_entry->offset);)
    return true;
  }/*}}}*/
  return false;
}/*}}}*/

// #define DEBUG_SDP_PDU_PREPARE

uint32_t SdpParser::prepare_sdp_pdu_message(uint8_t * dest, const uint8_t * src, uint32_t size) {/*{{{*/

  sdp_data_element_t sdp(size);
  uint8_t * des_token_pos      = const_cast<uint8_t *>(src);
  uint32_t prev_working_offset = 0;
  uint32_t total_length        = 0;
#ifdef DEBUG_SDP_PDU_PREPARE
  uint16_t prev_indent         = 0;

  for ( uint16_t pdu_offset = 0 ; pdu_offset < size ; pdu_offset++ ) {
    if (!(pdu_offset % 16)) printf("\r\n%c  ", pdu_offset == 0 ? '>' : ' ');
    if (!(pdu_offset % 8)) printf(" ");
    printf(" 0x%02.2x", des_token_pos[pdu_offset]);
  } 
  printf("... \r\n\r\n");
#endif

  while ( data_element_sequence_tokenizer( &sdp, &des_token_pos ) ) {
    if ( sdp.fragment_length > 0 ) {
      memmove(&(dest[prev_working_offset]), &(sdp.fragment[0]), sdp.fragment_length);
      total_length += sdp.fragment_length;
#ifdef DEBUG_SDP_PDU_PREPARE
      for ( uint32_t pdu_offset = 0 ; pdu_offset < sdp.fragment_length ; pdu_offset++ ) {
        if (!(pdu_offset % 16)) {
          printf("\r\n0x%04.4X/0x%04.4X % 4d ", 
            prev_working_offset,
            sdp.parse_limit,
            sdp.parse_depth > 0 ? sdp.container_sizes[sdp.parse_depth].size : sdp.fragment_length
          );
          for ( uint16_t indent = 0 ; indent < prev_indent ; indent++ ) {
            printf("  ");
          }
        }
        if (!(pdu_offset % 8)) printf(" ");
        printf(" 0x%02.2x", sdp.fragment[pdu_offset]);
      } 
#endif
    }
    if ( sdp.complete ) {
      size_offset_t & s = sdp.get_typedesc0(); 
      sdp.update_typespec( dest, &s );
    }
    prev_working_offset = sdp.working_offset;
#ifdef DEBUG_SDP_PDU_PREPARE
    prev_indent = sdp.parse_depth;
#endif
  }

#ifdef DEBUG_SDP_PDU_PREPARE
  printf("\r\n");
  for ( uint16_t pdu_offset = 0 ; pdu_offset < total_length; pdu_offset++ ) {
    if (!(pdu_offset % 16)) printf("\r\n%c  ", pdu_offset == 0 ? '>' : ' ');
    if (!(pdu_offset % 8)) printf(" ");
    printf(" 0x%02.2x", dest[pdu_offset]);
  } 
  printf(" ...\r\n\r\n");
#endif

  if ( sdp.parse_limit > 0 && sdp.working_offset < sdp.parse_limit ) {
    return 0;
  }

#ifdef DEBUG_SDP_PDU_PREPARE
  printf("* Final parsed data length: %lu\r\n\r\n", total_length);
#endif
  return total_length;

}/*}}}*/
