/* Copyright (C) 2012 Kristian Lauszus, TKJ Electronics. All rights reserved.

   This software may be distributed and modified under the terms of the GNU
   General Public License version 2 (GPL2) as published by the Free Software
   Foundation and appearing in the file GPL2.TXT included in the packaging of
   this file. Please note that GPL2 Section 2[b] requires that all works based
   on this software must also be made publicly available under the terms of
   the GPL2 ("Copyleft").

   Contact information
   -------------------

   Kristian Lauszus, TKJ Electronics
Web      :  http://www.tkjelectronics.com
e-mail   :  kristianl@tkjelectronics.com
*/

#ifndef _spp_h_
#define _spp_h_

#include "BTD.h"

#ifdef SPP_C_CODE
#define   SPP_CXX_IMPL(x) x
#define   SPP_VIRT_IMPL(x) SPP::x
#else
#define   SPP_CXX_IMPL(x) SPP::x
#define   SPP_VIRT_IMPL(x) SPP::x
#endif

/** This BluetoothService class implements the Serial Port Protocol (SPP). */
class SPP : public BluetoothService {
  public:
    /**
     * Constructor for the SPP class.
     * @param  p   Pointer to BTD class instance.
     * @param  name   Set the name to BTD#btdName. If argument is omitted, then "Arduino" will be used.
     * @param  pin   Write the pin to BTD#btdPin. If argument is omitted, then "1234" will be used.
     */
    SPP(BTD *p);

    /**
     * @return Return true if SPP communication is connected.
     */
    operator bool();

    /** @name BluetoothService implementation */
    /**
     * Used to pass acldata to the services.
     */
    virtual void ACLData(uint8_t *l2capinbuf);
    /**
     *  Used to establish the connection automatically.
     *  @return True if deferred command queue is nonempty
     */

    virtual bool Run(uint8_t *l2capinbuf);
    /** Use this to reset the service. */
    virtual void Reset();
    /** Update internal state of class instance */
    virtual void StateChange(uint32_t);
    /** Used this to disconnect the virtual serial port. */
    virtual void disconnect();
    /**@}*/

    virtual uint8_t getServiceId();

    // Class initialization
    virtual void Initialize(BTD *p);

    /** @name Serial port profile (SPP) Print functions */
    /**
     * Get number of bytes waiting to be read.
     * @return Return the number of bytes ready to be read.
     */
    virtual int available(void);
    /** Discard all the bytes in the buffer. */
    virtual void flush(void);
    /**
     * Used to read the next value in the buffer without advancing to the next one.
     * @return Return the byte. Will return -1 if no bytes are available.
     */
    virtual int peek(void);
    /**
     * Used to read the buffer.
     * @return Return the byte. Will return -1 if no bytes are available.
     */
    virtual int read(void);
    /**
     * Writes the byte to send to a buffer. The message is send when either send() or after Usb.Task() is called.
     * @param  data The byte to write.
     * @return      Return the number of bytes written.
     */
    virtual size_t write(uint8_t data);
    /**
     * Writes the bytes to send to a buffer. The message is send when either send() or after Usb.Task() is called.
     * @param  data The data array to send.
     * @param  size Size of the data.
     * @return      Return the number of bytes written.
     */
    virtual size_t write(const uint8_t* data, size_t size);
    /** Pull in write(const char *str) from Print */
    /**
     * This will send all the bytes in the buffer.
     * This is called whenever Usb.Task() is called,
     * but can also be called via this function.
     */
    void send(void);
    /**@}*/

  private:

    void SDP_task(); // SDP state machine
    void RFCOMM_task(); // RFCOMM state machine

    // 2nd-level, per-channel dispatch handlers
    bool handle_signaling_channel(uint8_t *l2capinbuf);
    bool handle_sdp_channel(uint8_t *l2capinbuf);
    bool handle_rfcomm_channel(uint8_t *l2capinbuf);

  private:
    /* Bluetooth dongle library pointer */
    BTD *pBtd;
    uint8_t service_class_id;

    // Declarations that are only active when compiling in CXX mode 

    bool connected;
    /* Set true when a channel is created */
    bool SDPConnected;
    bool RFCOMMConnected;

    uint16_t hci_handle; // The HCI Handle for the connection

    /* Variables used by L2CAP state machines */
    uint8_t l2cap_sdp_state;
    uint8_t l2cap_rfcomm_state;
    uint16_t l2cap_event_flag; // l2cap flags of received Bluetooth events

    /* L2CAP Channels */
    uint8_t sdp_scid[2]; // L2CAP source CID for SDP
    uint8_t sdp_dcid[2]; // 0x0050
    uint8_t rfcomm_scid[2]; // L2CAP source CID for RFCOMM
    uint8_t rfcomm_dcid[2]; // 0x0051
    uint8_t identifier; // Identifier for command

    /* RFCOMM Variables */
    uint8_t rfcommChannel;
    uint8_t rfcommChannelConnection; // This is the channel the SPP channel will be running at
    uint8_t rfcommDirection;
    uint8_t rfcommCommandResponse;
    uint8_t rfcommFrameType;
    uint8_t rfcommPfBit;

    uint32_t timer;
    bool waitForLastCommand;
    bool creditSent;

    uint8_t rfcommbuf[10]; // Buffer for RFCOMM Commands
    uint8_t rfcommDataBuffer[100]; // Create a 100 sized buffer for incoming data
    uint8_t sppOutputBuffer[100]; // Create a 100 sized buffer for outgoing SPP data
    uint8_t spp_available_out;
    uint8_t rfcommAvailable;

    bool firstMessage; // Used to see if it's the first SDP request received
    uint8_t bytesRead; // Counter to see when it's time to send more credit

};


#endif // _spp_h_
